# QuESt
# version: 1.2
# 
# by
# Sandia National Laboratories
# 
# Copyright 2018 National Technology & Engineering Solutions of Sandia, LLC (NTESS). Under the terms of Contract DE-NA0003525 with NTESS, the U.S. Government retains certain rights in this software.

# NOTICE:
#
# For five (5) years from 8/16/2018 the United States Government is granted for itself and others acting on its behalf a paid-up, nonexclusive, irrevocable worldwide license in this data to reproduce, prepare derivative works, and perform publicly and display publicly, by or on behalf of the Government. There is provision for the possible extension of the term of this license. Subsequent to that period or any extension granted, the United States Government is granted for itself and others acting on its behalf a paid-up, nonexclusive, irrevocable worldwide license in this data to reproduce, prepare derivative works, distribute copies to the public, perform publicly and display publicly, and to permit others to do so. The specific term of the license can be identified by inquiry made to National Technology and Engineering Solutions of Sandia, LLC or DOE.
#
# NEITHER THE UNITED STATES GOVERNMENT, NOR THE UNITED STATES DEPARTMENT OF ENERGY, NOR NATIONAL TECHNOLOGY AND ENGINEERING SOLUTIONS OF SANDIA, LLC, NOR ANY OF THEIR EMPLOYEES, MAKES ANY WARRANTY, EXPRESS OR IMPLIED, OR ASSUMES ANY LEGAL RESPONSIBILITY FOR THE ACCURACY, COMPLETENESS, OR USEFULNESS OF ANY INFORMATION, APPARATUS, PRODUCT, OR PROCESS DISCLOSED, OR REPRESENTS THAT ITS USE WOULD NOT INFRINGE PRIVATELY OWNED RIGHTS.
# 
# Any licensee of this software has the obligation and responsibility to abide by the applicable export control laws, regulations, and general prohibitions relating to the export of technical data. Failure to obtain an export control license or other authority from the Government may result in criminal liability under U.S. laws.


from __future__ import absolute_import

# This is for setting the window parameters like the initial size. Goes before any other import statements.
from kivy.config import Config, ConfigParser

Config.set('graphics', 'height', '900')
Config.set('graphics', 'width', '1600')
Config.set('graphics', 'minimum_height', '900')
Config.set('graphics', 'minimum_width', '1600')
#Config.set('graphics', 'borderless', '1')
Config.set('graphics', 'resizable', '1')
#Config.set('graphics', 'fullscreen', 'auto')
Config.set('kivy', 'desktop', 1)
Config.set('kivy', 'exit_on_escape', '0')  # disables Esc to quit
Config.set('input', 'mouse', 'mouse,multitouch_on_demand')  # disables red dot creation

from functools import partial
import os
import webbrowser
import threading

from kivy.utils import get_color_from_hex
from kivy.lang import Builder
from kivy.app import App
from kivy.clock import Clock
from kivy.uix.screenmanager import ScreenManager, Screen, RiseInTransition, SwapTransition
from kivy.uix.button import Button
from kivy.uix.modalview import ModalView
from kivy.uix.popup import Popup
from kivy.uix.boxlayout import BoxLayout
from kivy.uix.actionbar import ActionBar, ActionGroup
from kivy.properties import ObjectProperty
from kivy.core.text import LabelBase

from es_gui.apps.data_manager.data_manager import DataManager
from es_gui.resources.widgets.common import MyPopup, WarningPopup, APP_NAME, APP_TAGLINE, NavigationButton
from es_gui.proving_grounds.help_carousel import HelpCarouselModalView

dirname = os.path.dirname(__file__)

# Import common widgets from look_and_feel
Builder.load_file(os.path.join(dirname, 'es_gui', 'resources', 'widgets', 'common.kv'))

from es_gui.settings import ESAppSettings

# Data Manager
from es_gui.apps.data_manager.home import DataManagerHomeScreen
from es_gui.apps.data_manager.widgets import DataManagerRTOMOdataScreen
from es_gui.apps.data_manager.rate_structure import RateStructureDataScreen
from es_gui.apps.data_manager.load import DataManagerLoadHomeScreen, DataManagerCommercialLoadScreen, DataManagerResidentialLoadScreen
from es_gui.apps.data_manager.pv import PVwattsSearchScreen
from es_gui.apps.data_manager.nsrdb import NSRDBDataScreen

# Valuation
from es_gui.apps.valuation.home import ValuationHomeScreen
from es_gui.apps.valuation.batchrunscreen import BatchRunScreen
from es_gui.apps.valuation.results_viewer import ValuationResultsViewer
from es_gui.apps.valuation.wizard import ValuationWizard

# Behind-the-meter
from es_gui.apps.btm.home import BehindTheMeterHomeScreen
from es_gui.apps.btm.cost_savings import CostSavingsWizard
from es_gui.apps.btm.results_viewer import BtmResultsViewer

from es_gui.apps.performance.home import PerformanceHomeScreen
from es_gui.apps.performance.performance_sim import PerformanceSimRunScreen
from es_gui.apps.performance.results_viewer import PerformanceResultsViewer

# Technology selection
from es_gui.apps.tech_selection.home import TechSelectionHomeScreen
from es_gui.apps.tech_selection.tech_selection_wizard import TechSelectionWizard
from es_gui.apps.tech_selection.results_viewer import TechSelectionFeasible

# Font registration.
LabelBase.register(name='Exo 2',
                   fn_regular=os.path.join('es_gui', 'resources', 'fonts', 'Exo_2', 'Exo2-Regular.ttf'),
                   fn_bold=os.path.join('es_gui', 'resources', 'fonts', 'Exo_2', 'Exo2-Bold.ttf'),
                   fn_italic=os.path.join('es_gui', 'resources', 'fonts', 'Exo_2', 'Exo2-Italic.ttf'))

LabelBase.register(name='Open Sans',
                   fn_regular=os.path.join('es_gui', 'resources', 'fonts', 'Open_Sans', 'OpenSans-Regular.ttf'),
                   fn_bold=os.path.join('es_gui', 'resources', 'fonts', 'Open_Sans', 'OpenSans-Bold.ttf'),
                   fn_italic=os.path.join('es_gui', 'resources', 'fonts', 'Open_Sans', 'OpenSans-Italic.ttf'))

LabelBase.register(name='Modern Pictograms',
                   fn_regular=os.path.join('es_gui', 'resources', 'fonts', 'modernpictograms', 'ModernPictograms.ttf'))


class IndexScreen(Screen):
    """The landing screen."""
    def on_leave(self):
        """Sets NavigationBar.reset_nav_bar() to fire on_enter for the index screen after the first time loading it."""
        self.bind(on_enter=self.manager.nav_bar.reset_nav_bar)
    
    def open_intro_help_carousel(self):
        """
        """
        help_carousel_view = HelpCarouselModalView()
        help_carousel_view.title.text = "Welcome to QuESt"

        slide_01_text = "QuESt is an application suite for energy storage valuation.\n\nThe list on the left contains the currently available applications. Click on an application to learn a little more about it. Once you have selected an application, click on the 'Get started' button underneath its description to open it."

        slide_02_text = "At the top of the QuESt window is the action bar. The QuESt logo on the left end of the action bar serves as a back button; click on it to return to the previous screen. On the right end of the action bar is the navigation toolbar. The buttons here change depending on the context but several, like those pictured, persist.\n\nYou can use the 'home' button to return to this index screen at any time."

        slide_03_text = "In QuESt, input data management is separate from the analysis tools. Use the QuESt Data Manager to acquire data before proceeding to other QuESt applications and using their analysis tools."

        slide_04_text = "In some QuESt applications, it is possible to import and use your own data. Look out for prompts such as these to open the data importer interface. Please refer to each individual application and tool for specific details!"

        slide_05_text = "Looking for more help? Check the navigation bar while in each QuESt application for a 'help' button to open an information carousel like this one for application-specific help."

        slides = [
            (os.path.join("es_gui", "resources", "help_views", "index", "01.png"), slide_01_text),
            (os.path.join("es_gui", "resources", "help_views", "index", "02.png"), slide_02_text),
            (os.path.join("es_gui", "resources", "help_views", "index", "03.png"), slide_03_text),
            (os.path.join("es_gui", "resources", "help_views", "index", "04.png"), slide_04_text),
            (os.path.join("es_gui", "resources", "help_views", "index", "05.png"), slide_05_text),
        ]

        help_carousel_view.add_slides(slides)
        help_carousel_view.open()


class AboutScreen(ModalView):
    """The about/contact screen."""
    def __init__(self, **kwargs):
        super(AboutScreen, self).__init__(**kwargs)

        def _ref_link(text, ref):
            return '[ref={0}][color=003359][u]{1}[/u][/color][/ref]'.format(ref, text)

        def _go_to_webpage(instance, value):
            if value == 'kivy':
                webbrowser.open('http://kivy.org/')
            elif value == 'pyomo':
                webbrowser.open('http://pyomo.org/')
            elif value == 'sandia-ess':
                webbrowser.open('http://energy.sandia.gov/energy/ssrei/energy-storage/')
            elif value == 'sandia-epsr':
                webbrowser.open('http://energy.sandia.gov/energy/ssrei/gridmod/transmission-planning-and-operations/')
            elif value == 'sandia':
                webbrowser.open('http://sandia.gov/')
        
        version_statement = 'QuESt v1.2.f \n 2020.01.17'

        developed_by = '{app_name} is developed by the {ess} and {espr} departments at {sandia}.'.format(app_name=APP_NAME, ess=_ref_link('Energy Storage Technology and Systems', 'sandia-ess'), espr=_ref_link('Electric Power Systems Research', 'sandia-espr'), sandia=_ref_link('Sandia National Laboratories', 'sandia'))

        powered_by = '{app_name} is powered by Kivy and Pyomo. {kivy_ref} is an open-source Python library for rapid development of applications that make use of innovative user interfaces, such as multi-touch apps. {pyomo_ref} is a Python-based open-source software package that supports a diverse set of optimization capabilities for formulating, solving, and analyzing optimization models.'.format(app_name=APP_NAME, kivy_ref=_ref_link('Kivy', 'kivy'), pyomo_ref=_ref_link('Pyomo', 'pyomo'))

        acknowledgement = 'The developers would like to thank [b]Dr. Imre Gyuk[/b] at the Energy Storage Program at the U.S. Department of Energy for funding the development of this software.'

        ntess_statement = 'Sandia National Laboratories is a multimission laboratory managed and operated by National Technology and Engineering Solutions of Sandia, LLC, a wholly owned subsidiary of Honeywell International, Inc., for the U.S. Department of Energy\'s National Nuclear Security Administration under contract DE-NA0003525.'

        copyright_statement = 'Copyright 2018 National Technology & Engineering Solutions of Sandia, LLC (NTESS). Under the terms of Contract DE-NA0003525 with NTESS, the U.S. Government retains certain rights in this software. \n\n NOTICE: \n\n For five (5) years from 8/16/2018 the United States Government is granted for itself and others acting on its behalf a paid-up, nonexclusive, irrevocable worldwide license in this data to reproduce, prepare derivative works, and perform publicly and display publicly, by or on behalf of the Government. There is provision for the possible extension of the term of this license. Subsequent to that period or any extension granted, the United States Government is granted for itself and others acting on its behalf a paid-up, nonexclusive, irrevocable worldwide license in this data to reproduce, prepare derivative works, distribute copies to the public, perform publicly and display publicly, and to permit others to do so. The specific term of the license can be identified by inquiry made to National Technology and Engineering Solutions of Sandia, LLC or DOE. \n\n NEITHER THE UNITED STATES GOVERNMENT, NOR THE UNITED STATES DEPARTMENT OF ENERGY, NOR NATIONAL TECHNOLOGY AND ENGINEERING SOLUTIONS OF SANDIA, LLC, NOR ANY OF THEIR EMPLOYEES, MAKES ANY WARRANTY, EXPRESS OR IMPLIED, OR ASSUMES ANY LEGAL RESPONSIBILITY FOR THE ACCURACY, COMPLETENESS, OR USEFULNESS OF ANY INFORMATION, APPARATUS, PRODUCT, OR PROCESS DISCLOSED, OR REPRESENTS THAT ITS USE WOULD NOT INFRINGE PRIVATELY OWNED RIGHTS. \n\n Any licensee of this software has the obligation and responsibility to abide by the applicable export control laws, regulations, and general prohibitions relating to the export of technical data. Failure to obtain an export control license or other authority from the Government may result in criminal liability under U.S. laws.'

        third_party_code_statement = 'This software uses the following open source components under their respective licenses. Portions of these components are the copyrighted material of their respective authors.'

        holidays_license = 'Holidays is used for determining when weekend rates apply for utility rate structures. \n\n Copyright (c) 2014-2017 <ryanssdev@icloud.com> \n Copyright (c) 2018 <maurizio.montel@gmail.com> \n\n Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions: \n\n The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software. \n\n THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.'

        jinja2_license = 'Jinja2 is used for creating auto-generated HTML reports. \n\n Copyright (c) 2009 by the Jinja Team, see AUTHORS for more details. \n\n Some rights reserved. \n\n Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met: \n\n * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer. \n\n * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution. \n\n * The names of the contributors may not be used to endorse or promote products derived from this software without specific prior written permission. \n\n THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.'

        kivy_license = 'Kivy is used to build the GUI. \n\n Copyright (c) 2010-2018 Kivy Team and other contributors \n\n Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions: \n\n The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software. \n\n THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.'

        kivy_garden_license = 'Kivy Garden is used to extend the capabilities of Kivy. \n\n Copyright (c) 2010-2014 Kivy Team and other contributors \n\n Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions: \n\n The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software. \n\n THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.'

        kivy_garden_matplotlib_license = 'Kivy Garden.Matplotlib is used to enable Matplotlib capabilities, such as plotting, within QuESt. \n\n The MIT License (MIT) \n\n Copyright (c) 2015 Kivy Garden \n\n Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions: \n\n The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software. \n\n THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.'

        matplotlib_license = 'Matplotlib is used to plot data in results viewers. \n\n Copyright (c) 2012-2013 Matplotlib Development Team; All Rights Reserved. \n\n License agreement for matplotlib 2.2.2 \n\n 1. This LICENSE AGREEMENT is between the Matplotlib Development Team (“MDT”), and the Individual or Organization (“Licensee”) accessing and otherwise using matplotlib software in source or binary form and its associated documentation. \n\n 2. Subject to the terms and conditions of this License Agreement, MDT hereby grants Licensee a nonexclusive, royalty-free, world-wide license to reproduce, analyze, test, perform and/or display publicly, prepare derivative works, distribute, and otherwise use matplotlib 2.2.2 alone or in any derivative version, provided, however, that MDT’s License Agreement and MDT’s notice of copyright, i.e., “Copyright (c) 2012-2013 Matplotlib Development Team; All Rights Reserved” are retained in matplotlib 2.2.2 alone or in any derivative version prepared by Licensee. \n\n 3. In the event Licensee prepares a derivative work that is based on or incorporates matplotlib 2.2.2 or any part thereof, and wants to make the derivative work available to others as provided herein, then Licensee hereby agrees to include in any such work a brief summary of the changes made to matplotlib 2.2.2. \n\n 4. MDT is making matplotlib 2.2.2 available to Licensee on an “AS IS” basis. MDT MAKES NO REPRESENTATIONS OR WARRANTIES, EXPRESS OR IMPLIED. BY WAY OF EXAMPLE, BUT NOT LIMITATION, MDT MAKES NO AND DISCLAIMS ANY REPRESENTATION OR WARRANTY OF MERCHANTABILITY OR FITNESS FOR ANY PARTICULAR PURPOSE OR THAT THE USE OF MATPLOTLIB 2.2.2 WILL NOT INFRINGE ANY THIRD PARTY RIGHTS. \n\n 5. MDT SHALL NOT BE LIABLE TO LICENSEE OR ANY OTHER USERS OF MATPLOTLIB 2.2.2 FOR ANY INCIDENTAL, SPECIAL, OR CONSEQUENTIAL DAMAGES OR LOSS AS A RESULT OF MODIFYING, DISTRIBUTING, OR OTHERWISE USING MATPLOTLIB 2.2.2, OR ANY DERIVATIVE THEREOF, EVEN IF ADVISED OF THE POSSIBILITY THEREOF. \n\n 6. This License Agreement will automatically terminate upon a material breach of its terms and conditions. \n\n 7. Nothing in this License Agreement shall be deemed to create any relationship of agency, partnership, or joint venture between MDT and Licensee. This License Agreement does not grant permission to use MDT trademarks or trade name in a trademark sense to endorse or promote products or services of Licensee, or any third party. \n\n 8. By copying, installing or otherwise using matplotlib 2.2.2, Licensee agrees to be bound by the terms and conditions of this License Agreement.'

        numpy_license = 'NumPy is used for array objects. \n\n Copyright © 2005-2018, NumPy Developers. \n All rights reserved. \n\n Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met: \n\n Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer. \n Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution. \n Neither the name of the NumPy Developers nor the names of any contributors may be used to endorse or promote products derived from this software without specific prior written permission. \n\n THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS “AS IS” AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.'

        pandas_license = 'Pandas is used for data processing capabilities. \n\n BSD 3-Clause License \n\n Copyright (c) 2008-2012, AQR Capital Management, LLC, Lambda Foundry, Inc. and PyData Development Team \n All rights reserved. \n\n Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met: \n\n * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer. \n\n * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution. \n\n * Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission. \n\n THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.'

        pyomo_license = 'Pyomo is used to construct mathematical programs and to interface with solvers. \n\n Copyright 2008 Sandia Corporation. Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation, the U.S. Government retains certain rights in this software. \n\n All rights reserved. \n\n Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met: \n\n Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer. \n\n Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution. \n\n Neither the name of the Sandia National Laboratories nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission. \n\n THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.'

        requests_license = 'Requests is used for making HTTP requests, namely in the Data Manager. \n\n Copyright 2018 Kenneth Reitz. All rights reserved.  Licensed under the Apache License, Version 2.0, you may not use this file except in compliance with the Apache License.  You may obtain a copy of the Apache License at http://www.apache.org/licenses/LICENSE-2.0.  Unless required by applicable law or agreed to in writing, software distributed under the Apache License is distributed on an “AS IS” BASIS, WITHOUT WARRENTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the Apache License for the specific language governing permissions and limitations under the Apache License.'

        scipy_license = 'SciPy is used for scientific and engineering data processing. \n\n Copyright © 2001, 2002 Enthought, Inc. \n All rights reserved. \n\n Copyright © 2003-2013 SciPy Developers. \n All rights reserved. \n\n Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met: \n\n Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer. \n Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution. \n Neither the name of Enthought nor the names of the SciPy Developers may be used to endorse or promote products derived from this software without specific prior written permission. \n\n THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.'

        six_license = 'Six is used for Python compatibility utilities. \n\n Copyright (c) 2010-2018 Benjamin Peterson \n\n Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions: \n\n The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software. \n\n THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.'

        xlrd_license = 'xlrd is used for handling Microsoft Excel (tm) files. \n\n There are two licenses associated with xlrd. This one relates to the bulk of the work done on the library:: \n\n Portions copyright © 2005-2009, Stephen John Machin, Lingfo Pty Ltd \n All rights reserved. \n\n Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met: \n\n 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer. \n\n 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution. \n\n 3. None of the names of Stephen John Machin, Lingfo Pty Ltd and any contributors may be used to endorse or promote products derived from this software without specific prior written permission. \n\n THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. \n\n This one covers some earlier work:: \n\n Copyright (c) 2001 David Giffin. \n All rights reserved. \n Based on the the Java version: Andrew Khan Copyright (c) 2000. \n\n Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met: \n\n 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer. \n 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution. \n 3. All advertising materials mentioning features or use of this software must display the following acknowledgment: \n "This product includes software developed by David Giffin <david@giffin.org>." \n 4. Redistributions of any form whatsoever must retain the following acknowledgment: \n "This product includes software developed by David Giffin <david@giffin.org>." \n\n THIS SOFTWARE IS PROVIDED BY DAVID GIFFIN ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL DAVID GIFFIN OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.'

        third_party_code_licenses = '\n\n====================\n'.join([third_party_code_statement, holidays_license, jinja2_license, kivy_license, kivy_garden_license, kivy_garden_matplotlib_license, numpy_license, pandas_license, pyomo_license, requests_license, scipy_license, six_license, xlrd_license])

        self.about_label.text = '\n\n'.join([version_statement,
        developed_by, 
        #powered_by, 
        acknowledgement, 
        #ntess_statement, 
        copyright_statement,  
        third_party_code_licenses])

        self.about_label.bind(on_ref_press=_go_to_webpage)


class SettingsScreen(ModalView):
    """The settings screen. Driven by a custom Settings panel."""
    def on_close(self):
        self.dismiss()

        return True


class QuEStScreenManager(ScreenManager):
    """The screen manager for the overall application."""
    nav_bar = ObjectProperty()
    help_popup = ObjectProperty()

    def __init__(self, **kwargs):
        super(QuEStScreenManager, self).__init__(**kwargs)

        # Add new screens here.
        self.add_widget(IndexScreen())
        self.about_screen = AboutScreen()
        self.settings_screen = SettingsScreen()

        # Data manager.
        self.add_widget(DataManagerHomeScreen(name='data_manager_home'))
        self.add_widget(DataManagerRTOMOdataScreen(name='data_manager_rto_mo_data'))
        self.add_widget(RateStructureDataScreen(name='data_manager_rate_structure_data'))
        self.add_widget(DataManagerLoadHomeScreen(name='data_manager_load_home'))
        self.add_widget(DataManagerCommercialLoadScreen(name='data_manager_commercial_load'))
        self.add_widget(DataManagerResidentialLoadScreen(name='data_manager_residential_load'))
        self.add_widget(PVwattsSearchScreen(name='data_manager_pvwatts'))
        self.add_widget(NSRDBDataScreen(name='data_manager_nsrdb'))

        # Energy storage valuation.
        self.add_widget(ValuationHomeScreen(name='valuation_home'))
        self.add_widget(BatchRunScreen(name='batch_run'))
        self.add_widget(ValuationResultsViewer(name='valuation_results_viewer'))
        self.add_widget(ValuationWizard(name='valuation_wizard'))

        # Behind-the-meter applications.
        self.add_widget(BehindTheMeterHomeScreen(name='btm_home'))
        self.add_widget(CostSavingsWizard(name='cost_savings_wizard'))
        self.add_widget(BtmResultsViewer(name='btm_results_viewer'))
        
        #Performance applications.
        self.add_widget(PerformanceHomeScreen(name='performance_home'))
        self.add_widget(PerformanceSimRunScreen(name='performance_sim'))
        self.add_widget(PerformanceResultsViewer(name='performance_results_viewer'))
        
        # Technology selection application
        self.add_widget(TechSelectionHomeScreen(name='tech_selection_home'))
        self.add_widget(TechSelectionWizard(name='tech_selection_wizard'))
        self.add_widget(TechSelectionFeasible(name='feasible_techs'))
    
    def launch_valuation(self):
        """"""
        data_manager = App.get_running_app().data_manager

        try:
            data_manager.scan_valuation_data_bank()
        except FileNotFoundError:
            # 'data' directory does not exist.
            no_data_popup = WarningPopup()
            no_data_popup.popup_text.text = "Looks like you haven't downloaded any data yet. Try using QuESt Data Manager to get some data before returning here!"
            no_data_popup.open()
        else: 
            self.current = 'valuation_home'
    
    def launch_btm(self):
        """"""
        data_manager = App.get_running_app().data_manager

        try:
            data_manager.scan_btm_data_bank()
        except FileNotFoundError:
            # 'data' directory does not exist.
            no_data_popup = WarningPopup()
            no_data_popup.popup_text.text = "Looks like you haven't downloaded any data yet. Try using QuESt Data Manager to get some data before returning here!"
            no_data_popup.open()
        else: 
            self.current = 'btm_home'
            
    def launch_performance(self):
        """"""
        data_manager = App.get_running_app().data_manager
        
        try:
            data_manager.scan_performance_data_bank()
        except FileNotFoundError:
            # 'data' directory does not exist.
            no_data_popup = WarningPopup()
            no_data_popup.popup_text.text = "Looks like you haven't downloaded any data yet. Try using QuESt Data Manager to get some data before returning here!"
            no_data_popup.open()
        else: 
            self.current = 'performance_home'
            
    def launch_tech_selection(self):
        """"""    
        self.current = 'tech_selection_home'


class NavigationBar(ActionBar):
    """The dynamically updating navigation bar for traversing around the application."""
    parent_screen = {'index': 'index',
                     'valuation_home': 'index',
                     'batch_run': 'valuation_home',
                     'valuation_results_viewer': 'valuation_home',
                     'valuation_wizard': 'valuation_home',
                     'valuation_advanced': 'valuation_home',
                     'data_manager_home': 'index',
                     'data_manager_rto_mo_data': 'data_manager_home',
                     'data_manager_rate_structure_data': 'data_manager_home',
                     'data_manager_load_home': 'data_manager_home',
                     'data_manager_commercial_load': 'data_manager_load_home',
                     'data_manager_residential_load': 'data_manager_load_home',
                     'data_manager_pvwatts': 'data_manager_home',
                     'btm_home': 'index',
                     'cost_savings_wizard': 'btm_home',
                     'btm_results_viewer': 'btm_home',
                     'performance_home': 'index',
                     'performance_sim': 'performance_home',
                     'performance_results_viewer': 'performance_home',
                     'tech_selection_home': 'index',
                     'tech_selection_wizard': 'tech_selection_home',
                     }

    def __init__(self, sm):
        super(NavigationBar, self).__init__()
        self.sm = sm

        self.build_index_nav_bar()
    
    def build_data_manager_nav_bar(self):
        """Builds the navigation bar for data manager appliations."""
        data_manager_home_button = NavigationButton(
            text='data manager home',
            on_release=partial(self.go_to_screen, 'data_manager_home'),
        )

        self.reset_nav_bar()

        self.action_view.add_widget(data_manager_home_button)

    def build_valuation_advanced_nav_bar(self):
        """Builds the navigation bar for valuation applications."""
        view_results_button = NavigationButton(
            text='view results',
            on_release=partial(self.go_to_screen, 'valuation_results_viewer'),
        )

        run_op_button = NavigationButton(
            text='run optimization',
            on_release=self.sm.get_screen('set_parameters').execute_single_run,
        )

        load_data_button = NavigationButton(
            text='select data',
            on_release=partial(self.go_to_screen, 'load_data'),
        )

        set_parameters_button = NavigationButton(
            text='set parameters',
            on_release=partial(self.go_to_screen, 'set_parameters'),
        )

        self.reset_nav_bar()

        self.action_view.add_widget(load_data_button)
        self.action_view.add_widget(set_parameters_button)
        self.action_view.add_widget(run_op_button)
        self.action_view.add_widget(view_results_button)

    def build_valuation_results_nav_bar(self):
        """
        Builds the navigation bar for viewing results in valuation applications.
        """
        view_results_button = NavigationButton(
            text='view results',
            on_release=partial(self.go_to_screen, 'valuation_results_viewer'),
        )

        batch_processing_button = NavigationButton(
            text='batch runs',
            on_release=partial(self.go_to_screen, 'batch_run'),
        )

        self.reset_nav_bar()

        self.action_view.add_widget(view_results_button)
        self.action_view.add_widget(batch_processing_button)

    def build_valuation_batch_nav_bar(self):
        """Builds the navigation bar for batch processing in valuation applications."""
        view_results_button = NavigationButton(
            text='view results',
            on_release=partial(self.go_to_screen, 'valuation_results_viewer'),
        )

        batch_processing_button = NavigationButton(
            text='batch runs',
            on_release=partial(self.go_to_screen, 'batch_run'),
        )

        self.reset_nav_bar()

        self.action_view.add_widget(view_results_button)
        self.action_view.add_widget(batch_processing_button)
        
    def build_performance_nav_bar(self):
        """Builds the navigation bar for performance application"""
        performance_home_button = NavigationButton(
                text = 'performance home',
                on_release=partial(self.go_to_screen,'performance_home'))
        
        
        
        self.reset_nav_bar()

        self.action_view.add_widget(performance_home_button)

    def reset_nav_bar(self, *args):
        """Resets the navigation bar to its initial state."""
        # remove navigation buttons
        while len(self.action_view.children) > 1:
            for widget in self.action_view.children:
                if isinstance(widget, NavigationButton):
                    self.action_view.remove_widget(widget)

        # add home button and change title to 'Index'
        self.build_index_nav_bar()
        self.set_title('')

    def build_index_nav_bar(self):
        """Adds a home button to the navigation bar."""
        home_button = NavigationButton(
            text='home',
            on_release=partial(self.go_to_screen, 'index'),
        )

        settings_button = NavigationButton(
            text='settings',
            on_release=self.sm.settings_screen.open,
        )

        about_button = NavigationButton(
            text='about',
            on_release=self.sm.about_screen.open,
        )

        self.action_view.add_widget(home_button)
        self.action_view.add_widget(about_button)
        self.action_view.add_widget(settings_button)

    def go_to_screen(self, screen_name, *args):
        """Changes the current screen of the given screen manager."""
        self.sm.current = screen_name
    
    def go_up_screen(self):
        try:
            self.go_to_screen(self.parent_screen[self.sm.current])
        except:
            self.go_to_screen('index')

    def set_title(self, title):
        """Sets the title of the navigation bar."""
        self.action_view.action_previous.title = title


class QuEStApp(App):
    """
    The App class for launching the application.
    """
    config = ConfigParser()
    settings = ESAppSettings()

    def build_config(self, config):
        """Set default settings here."""
        config.setdefaults('optimization', {'solver': 'glpk'})
        config.setdefaults('connectivity', {'use_proxy': 0, 'http_proxy': '', 'https_proxy': '', 'use_ssl_verify': 1})
        config.setdefaults('valuation', {'valuation_dms_save': 1, 'valuation_dms_size': 20000})
        config.setdefaults('btm', {'btm_dms_save': 1, 'btm_dms_size': 20000})
        config.setdefaults('datamanager-pjm', {'pjm_subscription_key': ''})
        config.setdefaults('datamanager-isone', {'iso-ne_api_username': ''})
        config.setdefaults('datamanager-openei', {'openei_key': ''})
        config.setdefaults('performance', {'performance_dms_save': 1, 'performance_dms_size': 20000})

    def build(self):
        # Sets the window/application title.
        self.title = APP_NAME
        self.icon = 'es_gui/resources/logo/Quest_App_Icon_256.png'

        # Create ScreenManager.
        sm = QuEStScreenManager(transition=RiseInTransition(duration=0.2, clearcolor=[1, 1, 1, 1]))
        #sm = QuEStScreenManager(transition=SwapTransition(duration=0.2))

        # Instantiate DataManager.
        self.data_manager = DataManager()

        # Create BoxLayout container.
        bx = BoxLayout(orientation='vertical')

        # Add stop flag for threading management.
        bx.stop = threading.Event()

        # Create ActionBar and pass it a reference to the screen manager.
        ab = NavigationBar(sm)
        ab.sm = sm

        # Fill BoxLayout.
        bx.add_widget(ab)
        bx.add_widget(sm)

        # Pass reference of navigation bar to screen manager.
        sm.nav_bar = ab

        # Create Settings widget and add to settings screen.
        with open(os.path.join(dirname, 'es_gui', 'resources', 'settings', 'general.json'), 'r') as settings_json:
            self.settings.add_json_panel('General', self.config, data=settings_json.read())
        
        with open(os.path.join(dirname, 'es_gui', 'resources', 'settings', 'data_manager.json'), 'r') as settings_json:
            self.settings.add_json_panel('QuESt Data Manager', self.config, data=settings_json.read())
        
        with open(os.path.join(dirname, 'es_gui', 'resources', 'settings', 'valuation.json'), 'r') as settings_json:
            self.settings.add_json_panel('QuESt Valuation', self.config, data=settings_json.read())
        
        with open(os.path.join(dirname, 'es_gui', 'resources', 'settings', 'btm.json'), 'r') as settings_json:
            self.settings.add_json_panel('QuESt BTM', self.config, data=settings_json.read())
            
        with open(os.path.join(dirname,'es_gui','resources','settings','performance.json'),'r') as settings_json:
            self.settings.add_json_panel('QuESt Performance',self.config,data=settings_json.read())

        self.settings.bind(on_close=sm.settings_screen.dismiss)
        sm.settings_screen.settings_box.add_widget(self.settings)

        return bx

    def on_start(self):
        pass
    
    def on_stop(self):
        # Signal that the app is about to close
        self.root.stop.set()

if __name__ == '__main__':
    from kivy.core.window import Window

    # Sets window background color
    Window.clearcolor = get_color_from_hex('#FFFFFF')

    QuEStApp().run()



from setuptools import setup, find_packages

DISTNAME = "quest_eras"
VERSION = "1.6.3"
EXTENSIONS = []
PYTHON_REQUIRES = ">=3.8"
DESCRIPTION = "Sandia National Laboratories application suite for energy storage analysis and evaluation tools."
LONG_DESCRIPTION = "Sandia National Laboratories application suite for energy storage analysis and evaluation tools."
AUTHOR = "Sandia National Laboratories"
MAINTAINER_EMAIL = "tunguy@sandia.gov"
LICENSE = "BSD 3-clause"
URL = "https://www.github.com/snl-quest/snl-quest"

setuptools_kwargs = {
    "scripts": [],
    "include_package_data": True,
    "install_requires": [
        "numpy",
        "scipy",
        "pandas>=0.24.2",
        "pyomo>=5.6",
        "matplotlib>=3.5.1",
        "kivy>=1.10.1",
        "kivy-garden",
        "xlrd",
        "six",
        "jinja2",
        "bs4",
        "requests",
        "urllib3",
        "holidays",
        "seaborn",
        "eppy",
        "openpyxl",
        "pyutilib",
    ],
}

setup(
    name=DISTNAME,
    version=VERSION,
    packages=find_packages() + ['.'],
    ext_modules=EXTENSIONS,
    python_requires=PYTHON_REQUIRES,
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    author=AUTHOR,
    maintainer_email=MAINTAINER_EMAIL,
    license=LICENSE,
    url=URL,
    **setuptools_kwargs
)



from random import sample, choice
from functools import partial
from collections import OrderedDict
import calendar
import os
import datetime
import webbrowser
import base64
from time import sleep

from jinja2 import Environment, FileSystemLoader, select_autoescape
from kivy.uix.screenmanager import Screen, ScreenManager, WipeTransition, NoTransition
from kivy.uix.button import Button
from kivy.uix.togglebutton import ToggleButton
from kivy.uix.modalview import ModalView
from kivy.clock import Clock

from es_gui.proving_grounds.charts import BarChart, StackedBarChart, MultisetBarChart, PieChart, DonutChart, format_dollar_string
from es_gui.resources.widgets.common import TWO_ABC_WIDTH, THREE_ABC_WIDTH, MyPopup, TileButton, PALETTE, rgba_to_fraction, ReportScreen, WizardReportInterface, ReportChartToggle


class BtmCostSavingsReport(WizardReportInterface):
    chart_types = OrderedDict({
        'Total bill': 'total_bill_bar',
        # 'Total bill (by source)': 'total_bill_stacked',
        'Total bill comparison': 'total_bill_comparison_multiset',
        'Demand charge comparison': 'demand_charge_comparison_multiset',
        'Energy charge comparison': 'energy_charge_comparison_multiset',
        'NEM comparison': 'nem_comparison_multiset',
        'Peak demand comparison': 'peak_demand_comparison_multiset',
        # 'Total savings': 'total_savings_donut',
                   })
    
    # Sometimes the charges/bills/savings can be negative so we can't use stacked bar or donut charts.

    def __init__(self, chart_data, report_attributes, **kwargs):
        super(BtmCostSavingsReport, self).__init__(**kwargs)

        self.chart_type = type
        self.chart_data = chart_data
        self.report_attributes = report_attributes

        sm = self.report_sm
        # self.generate_report_button.disabled = True

        # Build chart type selection buttons and corresponding report screens.
        for opt in self.chart_types.items():
            button = ReportChartToggle(text=opt[0])
            button.bind(state=partial(self.add_report, opt[1]))
            self.chart_type_toggle.add_widget(button)

            screen = BtmCostSavingsReportScreen(type=opt[1], chart_data=self.chart_data, name=opt[1])
            sm.add_widget(screen)
    
    def has_chart(self, name):
        """Returns True if chart_data will generate a chart."""
        if name in {'demand_charge_comparison_multiset'}:
            has_chart_status = any(op[1].has_demand_charges() for op in self.chart_data)
        elif name in {'energy_charge_comparison_multiset'}:
            has_chart_status = any(op[1].has_energy_charges() for op in self.chart_data)
        elif name in {'nem_comparison_multiset'}:
            has_chart_status = any(op[1].has_nem_charges() for op in self.chart_data)
        else:
            has_chart_status = True
        
        return has_chart_status

    def add_report(self, chart_type, *args):
        # Adds a ReportScreen of type chart_type to the screen manager.
        sm = self.report_sm
        sm.transition = WipeTransition(duration=0.8, clearcolor=[1, 1, 1, 1])

        if not sm.has_screen(chart_type):
            screen = BtmCostSavingsReportScreen(type=chart_type, chart_data=self.chart_data, name=chart_type)
            sm.add_widget(screen)

        sm.current = chart_type
    
    def open_generate_report_menu(self):
        BtmCostSavingsGenerateReportMenu.host_report = self
        gen_report_menu = BtmCostSavingsGenerateReportMenu()
        gen_report_menu.open()


class ReportScreenManager(ScreenManager):
    transition = WipeTransition(duration=0.8, clearcolor=[1, 1, 1, 1])


class BtmCostSavingsReportScreen(ReportScreen):
    """A report screen for the BTM Cost Savings Wizard."""

    def __init__(self, type, chart_data, do_animation=True, **kwargs):
        super(BtmCostSavingsReportScreen, self).__init__(**kwargs)

        self.chart_type = type
        self.chart_data = chart_data
        self.do_animation = do_animation

        if self.chart_type == 'total_bill_bar':
            self.chart_bx.orientation = 'vertical'

            #self.desc.width = THREE_ABC_WIDTH
            self.desc_bx.size_hint_y = 0.33

            self.chart = BarChart(bar_spacing=25, y_axis_format='${0:,.0f}', size_hint_x=1.0, do_animation=self.do_animation)
            self.bind(on_enter=self.generate_total_bill_bar_chart)
        elif self.chart_type == 'total_bill_comparison_multiset':
            self.chart_bx.orientation = 'vertical'

            #self.desc.width = THREE_ABC_WIDTH
            self.desc_bx.size_hint_y = 0.33

            self.chart = MultisetBarChart(bar_spacing=25, legend_width=TWO_ABC_WIDTH/4, y_axis_format='${0:,.0f}', size_hint_x=1.0, do_animation=self.do_animation)
            self.bind(on_enter=self.generate_total_bill_comparison_chart)
        elif self.chart_type == 'demand_charge_comparison_multiset':
            self.chart_bx.orientation = 'vertical'

            #self.desc.width = THREE_ABC_WIDTH
            self.desc_bx.size_hint_y = 0.33

            self.chart = MultisetBarChart(bar_spacing=25, legend_width=TWO_ABC_WIDTH/4, y_axis_format='${0:,.0f}', size_hint_x=1.0, do_animation=self.do_animation)
            self.bind(on_enter=self.generate_demand_charge_comparison_chart)
        elif self.chart_type == 'energy_charge_comparison_multiset':
            self.chart_bx.orientation = 'vertical'

            #self.desc.width = THREE_ABC_WIDTH
            self.desc_bx.size_hint_y = 0.33

            self.chart = MultisetBarChart(bar_spacing=25, legend_width=TWO_ABC_WIDTH/4, y_axis_format='${0:,.0f}', size_hint_x=1.0, do_animation=self.do_animation)
            self.bind(on_enter=self.generate_energy_charge_comparison_chart)
        elif self.chart_type == 'nem_comparison_multiset':
            self.chart_bx.orientation = 'vertical'

            #self.desc.width = THREE_ABC_WIDTH
            self.desc_bx.size_hint_y = 0.33

            self.chart = MultisetBarChart(bar_spacing=25, legend_width=TWO_ABC_WIDTH/4, y_axis_format='${0:,.0f}', size_hint_x=1.0, do_animation=self.do_animation)
            self.bind(on_enter=self.generate_nem_charge_comparison_chart)
        elif self.chart_type == 'peak_demand_comparison_multiset':
            self.chart_bx.orientation = 'vertical'

            #self.desc.width = THREE_ABC_WIDTH
            self.desc_bx.size_hint_y = 0.33

            self.chart = MultisetBarChart(bar_spacing=25, legend_width=TWO_ABC_WIDTH/4, y_axis_format='{0:n} kW', size_hint_x=1.0, do_animation=self.do_animation)
            self.bind(on_enter=self.generate_peak_demand_comparison_chart)
        elif self.chart_type == 'total_savings_donut':
            self.chart_bx.orientation = 'vertical'

            #self.desc.width = THREE_ABC_WIDTH
            self.desc_bx.size_hint_y = 0.66

            self.chart = DonutChart(do_animation=self.do_animation)
            self.bind(on_enter=self.generate_total_savings_donut_chart)
        elif self.chart_type == 'total_bill_stacked':
            self.chart_bx.orientation = 'vertical'

            #self.desc.width = THREE_ABC_WIDTH
            self.desc_bx.size_hint_y = 0.33

            self.chart = StackedBarChart(bar_spacing=25, legend_width=TWO_ABC_WIDTH/4, y_axis_format='${0:,.0f}', size_hint_x=1.0, do_animation=self.do_animation)
            self.bind(on_enter=self.generate_total_bill_stackedbar_chart)
        else:
            raise(ValueError('An improper chart type was specified. (got {0})'.format(self.chart_type)))

        self.chart_bx.add_widget(self.chart)

    def on_leave(self):
        # reset the chart
        self.chart.clear_widgets()

    def generate_total_bill_bar_chart(self, *args):
        """Generates bar chart showing the total bill with energy storage each month."""
        bar_data = []

        if len(self.chart_data) > len(PALETTE):
            colors = PALETTE
        else:
            colors = sample(PALETTE, len(self.chart_data))

        for ix, op in enumerate(self.chart_data, start=0):
            name = op[0]
            month = name.split(' | ')[1]
            label = month

            solved_op = op[1]
            results = solved_op.results

            bar_color = colors[divmod(ix, len(colors))[1]]
            bar_data.append([label, rgba_to_fraction(bar_color), float(solved_op.total_bill_with_es)])

        self.chart.draw_chart(bar_data)
        max_bar = self.chart.max_bar
        min_bar = self.chart.min_bar

        # generate report text
        self.title.text = "Here's the total bill for each month."
        self.desc.text = 'The total bill is the sum of demand charges, energy charges, and net metering charges. '

        total_charges_str = format_dollar_string(sum(op[1].total_bill_with_es for op in self.chart_data))
        month_high_str = format_dollar_string(max_bar.value)
        month_low_str = format_dollar_string(min_bar.value)

        report_templates = [
        'The total charges for the year was [b]{0}[/b].'.format(total_charges_str),
        'The highest bill for a month was [b]{0}[/b] in [b]{1}[/b].'.format(month_high_str, calendar.month_name[list(calendar.month_abbr).index(max_bar.name)]),
        'The lowest bill for a month was [b]{0}[/b] in [b]{1}[/b].'.format(month_low_str, calendar.month_name[list(calendar.month_abbr).index(min_bar.name)]),
        ]

        self.desc.text += ' '.join(report_templates)
        self.is_drawn = True

    def generate_total_bill_comparison_chart(self, *args):
        """Generates the multiset bar chart comparing total bill with and without energy storage each month."""
        n_cats = 2

        if n_cats > len(PALETTE):
            colors = PALETTE
        else:
            colors = sample(PALETTE, n_cats)

        multisetbar_data = OrderedDict()

        for ix, op in enumerate(self.chart_data, start=0):
            name = op[0]
            month = name.split(' | ')[1]
            label = month

            solved_op = op[1]

            bar_group = [['without ES', rgba_to_fraction(colors[0]), float(solved_op.total_bill_without_es)]]
            bar_group.append(['with ES', rgba_to_fraction(colors[1]), float(solved_op.total_bill_with_es)])

            multisetbar_data[label] = bar_group

        self.chart.draw_chart(multisetbar_data)

        # generate report text
        self.title.text = "Here's the total bill with and without energy storage for each month."
        self.desc.text = 'The total bill is the sum of demand charges, energy charges, and net metering charges or credits. '

        report_templates = [
        ]

        total_bill_difference = sum(op[1].total_bill_with_es - op[1].total_bill_without_es for op in self.chart_data)

        if total_bill_difference < 0:
            relation = 'decrease'
            total_difference_str = "It looks like the ESS was able to [b]{0}[/b] the total charges over the year by [b]{1}[/b].".format(relation, format_dollar_string(abs(total_bill_difference)))
            report_templates.append(total_difference_str)

        self.desc.text += ' '.join(report_templates)
        self.is_drawn = True
    
    def generate_demand_charge_comparison_chart(self, *args):
        """Generates the multiset bar chart comparing total demand charges with and without energy storage each month."""
        if any(op[1].has_demand_charges() for op in self.chart_data):
            n_cats = 2

            if n_cats > len(PALETTE):
                colors = PALETTE
            else:
                colors = sample(PALETTE, n_cats)

            multisetbar_data = OrderedDict()

            for ix, op in enumerate(self.chart_data, start=0):
                name = op[0]
                month = name.split(' | ')[1]
                label = month

                solved_op = op[1]
                results = solved_op.results

                bar_group = [['without ES', rgba_to_fraction(colors[0]), float(solved_op.demand_charge_without_es)]]
                bar_group.append(['with ES', rgba_to_fraction(colors[1]), float(solved_op.demand_charge_with_es)])

                multisetbar_data[label] = bar_group

            self.chart.draw_chart(multisetbar_data)

            self.title.text = "Here are the demand charge totals each month."
            self.desc.text = 'The demand charge total consists of time-of-use peak demand charges in addition to a flat peak demand charge, if applicable. The time-of-use demand charge is based on the peak demand during each time period and the corresponding rate. The flat demand charge is based on the peak demand over the entire month, sometimes subject to minimum and/or maximum values. The ESS is useful for reducing net power draw during high time-of-use rates. '
            report_templates = [
            ]

            total_demand_charge_difference = sum(op[1].demand_charge_with_es - op[1].demand_charge_without_es for op in self.chart_data)

            if total_demand_charge_difference < 0:
                relation = 'decrease'
                total_difference_str = "It looks like the ESS was able to [b]{0}[/b] the total demand charges over the year by [b]{1}[/b].".format(relation, format_dollar_string(abs(total_demand_charge_difference)))
                report_templates.append(total_difference_str)

            self.desc.text += ' '.join(report_templates)
            self.is_drawn = True
        else:
            self.title.text = "It looks like there were no demand charges."
            self.desc.text = "The particular rate structure you selected resulted in no demand charges. Either there are no demand charges or no savings on demand charges were accrued using energy storage."

    def generate_energy_charge_comparison_chart(self, *args):
        """Generates the multiset bar chart comparing total energy charges with and without energy storage each month."""
        if any(op[1].has_energy_charges() for op in self.chart_data):
            n_cats = 2

            if n_cats > len(PALETTE):
                colors = PALETTE
            else:
                colors = sample(PALETTE, n_cats)

            multisetbar_data = OrderedDict()

            for ix, op in enumerate(self.chart_data, start=0):
                name = op[0]
                month = name.split(' | ')[1]
                label = month

                solved_op = op[1]
                results = solved_op.results

                bar_group = [['without ES', rgba_to_fraction(colors[0]), float(solved_op.energy_charge_without_es)]]
                bar_group.append(['with ES', rgba_to_fraction(colors[1]), float(solved_op.energy_charge_with_es)])

                multisetbar_data[label] = bar_group

            self.chart.draw_chart(multisetbar_data)

            self.title.text = "Here are the energy charge totals each month."
            self.desc.text = 'The energy charge total is based on net energy consumption and different time-of-use rates. The ESS is useful for reducing energy consumption during high time-of-use periods. '
            report_templates = [
            ]

            total_energy_charge_difference = sum(op[1].energy_charge_with_es - op[1].energy_charge_without_es for op in self.chart_data)

            if total_energy_charge_difference < 0:
                relation = 'decrease'
                total_difference_str = "It looks like the ESS was able to [b]{0}[/b] the total energy charges over the year by [b]{1}[/b].".format(relation, format_dollar_string(abs(total_energy_charge_difference)))
                report_templates.append(total_difference_str)
            else:
                relation = 'increased'
                total_difference_str = "It looks like the total energy charges over the year with the ESS [b]{0}[/b] by [b]{1}[/b]. This is likely due to opportunities for decreasing demand charges or obtaining net metering credits.".format(relation, format_dollar_string(abs(total_energy_charge_difference)))
                report_templates.append(total_difference_str)

            self.desc.text += ' '.join(report_templates)
            self.is_drawn = True
        else:
            self.title.text = "It looks like there were no energy charges."
            self.desc.text = "The particular rate structure you selected resulted in no energy charges. Either there are no time-of-use energy charges or no savings on energy charges were accrued using energy storage."
    
    def generate_nem_charge_comparison_chart(self, *args):
        """Generates the multiset bar chart comparing total net energy metering charges with and without energy storage each month."""
        if any(op[1].has_nem_charges() for op in self.chart_data):
            n_cats = 2

            if n_cats > len(PALETTE):
                colors = PALETTE
            else:
                colors = sample(PALETTE, n_cats)

            multisetbar_data = OrderedDict()

            for ix, op in enumerate(self.chart_data, start=0):
                name = op[0]
                month = name.split(' | ')[1]
                label = month

                solved_op = op[1]

                bar_group = [['without ES', rgba_to_fraction(colors[0]), float(solved_op.nem_charge_without_es)]]
                bar_group.append(['with ES', rgba_to_fraction(colors[1]), float(solved_op.nem_charge_with_es)])

                multisetbar_data[label] = bar_group
            
            self.chart.draw_chart(multisetbar_data)

            self.title.text = "Here are the net energy metering (NEM) totals each month."

            net_metering_type = self.chart_data[0][1].nem_type
            net_metering_rate = self.chart_data[0][1].nem_rate
            
            if net_metering_type == 2:
                net_metering_type_str = '[b]Net energy metering 2.0[/b] uses the time-of-use energy rate for energy.'
            else:
                net_metering_type_str = '[b]Net energy metering 1.0[/b] uses a fixed price for energy, which was [b]{0}[/b]/kWh.'.format(format_dollar_string(net_metering_rate))

            self.desc.text = "{0} Negative values represent credits. ".format(net_metering_type_str)
            report_templates = [
            ]

            total_nem_difference = sum(op[1].nem_charge_with_es - op[1].nem_charge_without_es for op in self.chart_data)

            relation = 'increase' if total_nem_difference < 0 else 'decrease'
            total_difference_str = "The total [b]{0}[/b] in NEM credits with energy storage was [b]{1}[/b].".format(relation, format_dollar_string(abs(total_nem_difference)))
            report_templates.append(total_difference_str)

            self.desc.text += ' '.join(report_templates)
            self.is_drawn = True
        else:
            self.title.text = "It looks like there were no net energy metering charges or credits."
            self.desc.text = "The particular rate structure you selected resulted in no net energy metering charges. Either that or no savings on net energy metering charges were accrued using energy storage."

    def generate_peak_demand_comparison_chart(self, *args):
        """Generates the multiset bar chart comparing peak monthly demand with and without energy storage each month."""
        n_cats = 2

        if n_cats > len(PALETTE):
            colors = PALETTE
        else:
            colors = sample(PALETTE, n_cats)

        multisetbar_data = OrderedDict()

        for ix, op in enumerate(self.chart_data, start=0):
            name = op[0]
            month = name.split(' | ')[1]
            label = month

            solved_op = op[1]
            results = solved_op.results

            pfpk_with_es = solved_op.model.pfpk.value
            pfpk_without_es = max(solved_op.model.pnet)

            bar_group = [['without ES', rgba_to_fraction(colors[0]), int(pfpk_without_es)]]
            bar_group.append(['with ES', rgba_to_fraction(colors[1]), int(pfpk_with_es)])

            multisetbar_data[label] = bar_group

        self.chart.draw_chart(multisetbar_data)

        self.title.text = "Here are the peak demand values each month."
        self.desc.text = 'The peak demand value each month is used to compute flat demand charges, if applicable. '
        report_templates = [
        ]

        if all(op[1].model.flt_dr == 0 for op in self.chart_data):
            report_templates.append("For this rate structure, there were no flat demand charges.")

        self.desc.text += ' '.join(report_templates)
        self.is_drawn = True

    def generate_total_savings_donut_chart(self, *args):
        """TODO: Probably deprecate this since we can't guarantee non-negative quantities."""
        donut_data = []
        savings_sources = [
            ('energy charges', 'energy_charge_with_es', 'energy_charge_without_es'),
            ('demand charges', 'demand_charge_with_es', 'demand_charge_without_es'),
            ('NEM charges', 'nem_charge_with_es', 'nem_demand_charge_without_es'),
        ]
        total_savings = 0

        if sum([x[1].nem_charge_without_es for x in self.chart_data]) + sum([x[1].nem_charge_with_es for x in self.chart_data]) <= 0:
            n_cats = 2
        else:
            n_cats = 3

        if n_cats > len(PALETTE):
            colors = PALETTE
        else:
            colors = sample(PALETTE, n_cats)

        for ix, source in enumerate(savings_sources[:n_cats], start=0):
            savings = sum([getattr(op[1], source[-1]) - getattr(op[1], source[1]) for op in self.chart_data])
            total_savings += savings

            donut_data.append([source[0], rgba_to_fraction(colors[ix]), savings])

        # generate chart
        self.chart.draw_chart(donut_data)

        # generate report text
        self.title.text = "Here are the savings from each charge source when using energy storage."
        self.desc.text = ''
        report_templates = [
            'The [b]total[/b] savings from energy storage was [b]${0:,.2f}[/b].'.format(total_savings),
        ]

        self.desc.text += ' '.join(report_templates)

    def generate_total_bill_stackedbar_chart(self, *args):
        """TODO: Probably deprecate this since we can't guarantee non-negative quantities."""
        stackedbar_data = OrderedDict()

        charge_sources = [
            ('energy charges', 'energy_charge_with_es'),
            ('demand charges', 'demand_charge_with_es'),
            ('NEM charges', 'nem_charge_with_es'),
        ]

        if sum([x[1].nem_charge_without_es for x in self.chart_data]) + sum([x[1].nem_charge_with_es for x in self.chart_data]) == 0:
            n_cats = 2
        else:
            n_cats = 3

        if n_cats > len(PALETTE):
            colors = PALETTE
        else:
            colors = sample(PALETTE, n_cats)

        for op in self.chart_data:
            name = op[0]
            month = name.split(' | ')[1]
            label = month

            solved_op = op[1]
            bar_stack = []

            for iy, source in enumerate(charge_sources[:n_cats], start=0):
                component_color = colors[iy]
                savings = float(getattr(solved_op, source[-1]))

                bar_stack.append([source[0], rgba_to_fraction(component_color), savings])

            stackedbar_data[label] = bar_stack

        self.chart.draw_chart(stackedbar_data)

        self.title.text = "Here's the breakdown of the total bill with energy storage each month."
        self.desc.text = ''

        report_templates = [
        ]

        self.desc.text += ' '.join(report_templates)


class BtmCostSavingsGenerateReportMenu(ModalView):
    host_report = None
    graphics_locations = {}
    report_id = None

    def __init__(self, **kwargs):
        super(BtmCostSavingsGenerateReportMenu, self).__init__(**kwargs)

        self.sm.transition = NoTransition()

        self.report_id = datetime.datetime.now().strftime('%Y_%m_%d_%H%M%S')

    def save_figure(self, screen, *args):
        if not self.sm.has_screen(screen.name):
            self.sm.add_widget(screen)

        self.sm.current = screen.name

        chart_images_dir = os.path.join('results', 'btm_cost_savings', 'report', self.report_id, 'images')
        os.makedirs(chart_images_dir, exist_ok=True)

        chart_save_location = os.path.join(chart_images_dir, 'chart_{n}.png'.format(n=screen.name))

        Clock.schedule_once(partial(screen.chart.export_to_png, chart_save_location), 1.0)

        # Save image name/path for report generator.
        self.graphics_locations[screen.name] = os.path.join('images', 'chart_{n}.png'.format(n=screen.name))

    def generate_report_screens(self):
        screen_flip_interval = 1.2
        n_charts = len(self.host_report.chart_types.items())

        # Draw figures for saving to .png.
        for ix, opt in enumerate(self.host_report.chart_types.items(), start=0):
            chart_name = opt[1]
            has_chart_status = self.host_report.has_chart(chart_name)
            
            if has_chart_status:
                screen = BtmCostSavingsReportScreen(type=chart_name, chart_data=self.host_report.chart_data, name=chart_name, do_animation=False)
                Clock.schedule_once(partial(self.save_figure, screen), ix * screen_flip_interval)

        self.generate_report_button.disabled = True

        # Generate report.
        Clock.schedule_once(lambda dt: self.generate_report_from_template(), (n_charts+1)*screen_flip_interval)
    
    def generate_executive_summary(self):
        """Generates an executive summary similar to the report screen using chart data."""
        chart_data = self.host_report.chart_data

        executive_summary_strings = []

        total_bill_summary = "The total bill is the sum of demand charges, energy charges, and net metering charges. The total charges for the year was <b>{total_charges}</b>. The highest bill for a month was <b>{highest_bill}</b>. The lowest bill for a month was <b>{lowest_bill}</b>.".format(
            total_charges=format_dollar_string(sum(op[1].total_bill_with_es for op in chart_data)),
            highest_bill=format_dollar_string(max(op[1].total_bill_with_es for op in chart_data)),
            lowest_bill=format_dollar_string(min(op[1].total_bill_with_es for op in chart_data)),
        )

        executive_summary_strings.append(total_bill_summary)

        bill_comparison_summary = "The total bill without energy storage was <b>{total_bill_without_es}</b>. The total bill with energy storage was <b>{total_bill_with_es}</b>: a difference of <b>{total_bill_diff}</b>.".format(
            total_bill_without_es=format_dollar_string(sum(op[1].total_bill_without_es for op in chart_data)),
            total_bill_with_es=format_dollar_string(sum(op[1].total_bill_with_es for op in chart_data)),
            total_bill_diff=format_dollar_string(abs(sum(op[1].total_bill_with_es - op[1].total_bill_without_es for op in chart_data))),
        )

        executive_summary_strings.append(bill_comparison_summary)

        demand_charge_strings = []
        demand_charge_strings.append("The demand charge total consists of time-of-use (TOU) peak demand charges in addition to flat peak demand charges, if applicable. The TOU demand charge is based on the peak demand during each time period and the corresponding rate. The flat demand charge is based on the peak demand over the entire month, sometimes subject to minimum and/or maximum values. The ESS is useful for reducing net power draw during high TOU rates.")

        demand_charge_summary = ' '.join(demand_charge_strings)
        executive_summary_strings.append(demand_charge_summary)
        demand_charge_strings = []

        peak_demand_without_es = max(max(op[1].model.pnet for op in chart_data))
        peak_demand_with_es = max(op[1].model.pfpk.value for op in chart_data)

        demand_charge_strings.append("Without energy storage, the peak demand observed during the evaluation period was <b>{peak_demand_without_es:.2f} kW</b>. By adding energy storage, this value was changed to <b>{peak_demand_with_es:.2f} kW</b>.".format(
            peak_demand_without_es=peak_demand_without_es,
            peak_demand_with_es=peak_demand_with_es,
        ))

        if any(op[1].has_demand_charges() for op in chart_data):
            demand_charge_with_es = sum(op[1].demand_charge_with_es for op in chart_data)
            demand_charge_without_es = sum(op[1].demand_charge_without_es for op in chart_data)
            total_demand_charge_difference = demand_charge_with_es - demand_charge_without_es

            if total_demand_charge_difference < 0:
                relation = 'decrease'
            else:
                relation = 'increase'
            
            demand_charge_strings.append(
                "It looks like the ESS was able to <b>{relation}</b> the total demand charges over the year from <b>{demand_charge_without_es}</b> to <b>{demand_charge_with_es}</b> for a total difference of <b>{total_demand_charge_difference}</b>.".format(
                    relation=relation,
                    demand_charge_without_es=format_dollar_string(demand_charge_without_es),
                    demand_charge_with_es=format_dollar_string(demand_charge_with_es),
                    total_demand_charge_difference=format_dollar_string(total_demand_charge_difference),
                )
            )
        else:
            demand_charge_strings.append(
                "It looks like there were no demand charges. The particular rate structure you selected resulted in no demand charges. Either there are no demand charges or no savings on demand charges were accrued using energy storage."
            )
        
        demand_charge_summary = ' '.join(demand_charge_strings)
        executive_summary_strings.append(demand_charge_summary)

        energy_charge_strings = []
        energy_charge_strings.append("The energy charge total is based on the net energy consumption and TOU rates. The ESS is useful for reducing energy consumption during high TOU periods.")

        if any(op[1].has_energy_charges for op in chart_data):
            energy_charge_with_es = sum(op[1].energy_charge_with_es for op in chart_data)
            energy_charge_without_es = sum(op[1].energy_charge_without_es for op in chart_data)
            total_energy_charge_difference = energy_charge_with_es - energy_charge_without_es

            if total_energy_charge_difference < 0:
                relation = 'decrease'
                total_difference_str = "It looks like the ESS was able to <b>{relation}</b> the total energy charges over the year by <b>{total_difference}</b> from <b>{energy_charge_without_es}</b> to <b>{energy_charge_with_es}</b>.".format(
                    relation=relation,
                    total_difference=format_dollar_string(total_energy_charge_difference),
                    energy_charge_without_es=format_dollar_string(energy_charge_without_es),
                    energy_charge_with_es=format_dollar_string(energy_charge_with_es),
                )
            else:
                relation = 'increased'
                total_difference_str = "It looks like the total energy charges over the year with the ESS <b>{relation}</b> by <b>{total_difference}</b> from <b>{energy_charge_without_es}</b> to <b>{energy_charge_with_es}</b>. This is likely due to opportunities for decreasing demand charges or obtaining net metering credits.".format(
                    relation=relation,
                    total_difference=format_dollar_string(total_energy_charge_difference),
                    energy_charge_without_es=format_dollar_string(energy_charge_without_es),
                    energy_charge_with_es=format_dollar_string(energy_charge_with_es),
                )
            
            energy_charge_strings.append(total_difference_str)
        else:
            energy_charge_strings.append("It looks like there were no energy charges. Either there are no TOU energy charges or no savings on energy charges were accrued using energy storage.")
        
        energy_charge_summary = ' '.join(energy_charge_strings)
        executive_summary_strings.append(energy_charge_summary)

        net_metering_strings = []
        net_metering_rate = chart_data[0][1].nem_rate
        net_metering_type = chart_data[0][1].nem_type

        if net_metering_type == 2:
            net_metering_strings.append("<b>Net energy metering (NEM) 2.0</b> uses the TOU energy rate for energy.")
        else:
            net_metering_strings.append("<b>Net energy metering (NEM) 1.0</b> uses a fixed price for energy, which was set as <b>{net_metering_rate}/kWh</b>.".format(net_metering_rate=format_dollar_string(net_metering_rate)))

        if any(op[1].has_nem_charges() for op in chart_data):
            net_metering_without_es = sum(op[1].nem_charge_without_es for op in chart_data)
            net_metering_with_es = sum(op[1].nem_charge_with_es for op in chart_data)
            total_net_metering_difference = net_metering_with_es - net_metering_without_es

            relation = 'increase' if total_net_metering_difference < 0 else 'decrease'

            net_metering_strings.append("The total <b>{relation}</b> in NEM credits with energy storage was <b>{total_nem_difference}</b>, from <b>{nem_without_es}</b> to <b>{nem_with_es}</b>.".format(
                relation=relation,
                total_nem_difference=format_dollar_string(total_net_metering_difference),
                nem_without_es=format_dollar_string(net_metering_without_es),
                nem_with_es=format_dollar_string(net_metering_with_es),
            ))
        else:
            net_metering_strings.append("It looks like there were no net energy metering charges or credits. Either that or no savings on NEM charges were accrued using energy storage.")
        
        net_metering_summary = ' '.join(net_metering_strings)
        executive_summary_strings.append(net_metering_summary)

        return executive_summary_strings

    def generate_report_from_template(self):
        # Get current date.
        now = datetime.datetime.now()
        today = now.strftime("%B %d, %Y")

        # Get report-specific data.
        op_handler_requests = self.host_report.report_attributes
        load_profile = op_handler_requests['load_profile']
        pv_profile = op_handler_requests['pv_profile']
        system_params = op_handler_requests['param desc']
        rate_structure = op_handler_requests['rate_structure']

        chart_types = self.host_report.chart_types
        chart_list = []
        chart_ix = 1

        for chart_caption, chart_type in chart_types.items():
            if self.host_report.has_chart(chart_type):
                chart_info = {}
                chart_info['path'] = self.graphics_locations[chart_type]
                chart_info['ix'] = chart_ix
                chart_info['caption'] = chart_caption

                chart_list.append(chart_info)
                chart_ix += 1

        utility = {'name': 'Utility', 'value': rate_structure['utility']['utility name']}
        utility_rate_structure = {'name': 'Rate Structure', 'value': rate_structure['utility']['rate structure name']}
        load_profile_name = {'name': 'Load Profile', 'value': load_profile['name']}

        summary_components = [
            utility,
            utility_rate_structure,
            load_profile_name
        ]

        pv_profile_summary = pv_profile.get('descriptors', [])

        system_parameters_summary = []

        for parameter in system_params:
            name, val_units = parameter.split(':')
            value, units = val_units.split()

            system_parameters_summary.append({'name': name, 'value': value, 'units': units})
        
        executive_summary = self.generate_executive_summary()

        template_dir = os.path.join('es_gui', 'resources', 'report_templates')
        output_dir_name = self.report_id

        output_dir = os.path.join('results', 'btm_cost_savings', 'report', output_dir_name)
        os.makedirs(output_dir, exist_ok=True)

        # Initialize Jinja environment.
        env = Environment(loader=FileSystemLoader(template_dir), autoescape=select_autoescape(['html']))
        template = env.get_template('btm_cost_savings.html')
        fname = os.path.join(output_dir, 'BTM_cost_savings_report.html')

        # Render output file.
        output = template.render(
                        # GENERAL OUTPUT
                        today=today,
                        header="This report shows the results from optimizations performed by QuESt BTM.",
                        summary_components=summary_components,
                        pv_profile_summary=pv_profile_summary,
                        system_parameters_summary=system_parameters_summary,   
                        QuESt_Logo=os.path.join('..', 'images', 'static', 'Quest_Logo_RGB.png'),
                        SNL_image=os.path.join('..', 'images', 'static', 'SNL.png'),
                        DOE_image=os.path.join('..', 'images', 'static', 'DOE.png'),
                        acknowledgement="Sandia National Laboratories is a multimission laboratory managed and operated by National Technology & Engineering Solutions of Sandia, LLC, a wholly owned subsidiary of Honeywell International Inc., for the U.S. Department of Energy's National Nuclear Security Administration under contract DE-NA0003525.",
                        executive_summary=executive_summary,					 
                        # FIGURES
                        charts=chart_list,
					)

        with open(fname,"w") as f:
            f.write(output)
        
        completion_popup = OpenGeneratedReportPopup()
        completion_popup.report_filename = fname
        completion_popup.open()


class OpenGeneratedReportPopup(MyPopup):
    report_filename = ''

    def open_generated_report(self):
        """Opens the generated report and dismisses the popup,"""
        webbrowser.open('file://' + os.path.realpath(self.report_filename))
        self.dismiss()


"""Data Management System."""

from __future__ import print_function, absolute_import

from collections import OrderedDict
import pickle
import logging
import os
import numpy as np


class DataManagementSystem:
    """
    A class used to store processed DataFrames as NumPy ndarrays and manage memory consumed.

    Data is stored in nested dictionaries up to a depth of 2: {key_0: {key_0_0: data}}.
    When the calculated memory exceeds max_memory, the dictionary at depth 1 at the front of the queue is popped out of the dictionary
    until the memory consumption is less than the maximum. The queue is determined by time of accessing. Accessing or adding to the
    structure at any depth will push the depth 1 dictionary to the back of the queue.

    :param save_name: The path/filename to pickle the DMS's data.
    :param max_memory: The maximum amount of memory, in bytes, that the contained ndarrays may collectively occupy.
    """

    def __init__(self, save_name, save_data=False, max_memory=500000):
        self.memory_used = 0
        self.max_memory = max_memory
        self.save_data = save_data
        self.save_name = save_name

        try:
            with open(self.save_name, "rb") as pfile:
                self.data = pickle.load(pfile)
                logging.info(
                    "DMS: Successfully loaded {fname}.".format(fname=self.save_name)
                )
        except FileNotFoundError:
            self.data = OrderedDict()
        except pickle.PickleError:
            logging.error("DMS: Could not unpickle data; purging and restarting DMS.")
            self.delete_pickle()
            self.data = OrderedDict()

    def delete_pickle(self):
        """Delete the pickle file used for self.data object persistence."""
        os.remove(self.save_data)

    def save_state(self):
        """Pickles self.data at self.save_name."""
        if self.save_data:
            logging.info("DMS: Saving {0}.".format(self.save_name))
            with open(self.save_name, "wb") as pfile:
                pickle.dump(self.data, pfile, protocol=3)

    def pop(self):
        """Shortcut for popping the queue of the OrderedDict."""
        k, v = self.data.popitem(last=False)
        print("Popped: ", (k))

    def requeue(self, key):
        """Move self.data[key] to the back of the queue for being purged."""
        if key in self.data:
            to_requeue = self.data.pop(key)
            self.data[key] = to_requeue

    def manage_memory(self):
        """Pop entries from the queue until occupied memory is less than the maximum allocated."""
        dms_sz = self.compute_memory()

        while dms_sz > self.max_memory:
            print("Memory limit exceeded. Purging old data...")
            print("Currently using: ", dms_sz, "bytes")
            print("Maximum allowed: ", self.max_memory, "bytes")
            self.pop()
            dms_sz = self.compute_memory()
            print("Now using: ", dms_sz, "bytes")

        self.save_state()

    def compute_memory(self):
        """Compute the memory footprint of the entire data structure."""

        def _compute_memory(coll):
            dms_sz = 0

            for value in coll.values():
                if isinstance(value, np.ndarray):
                    array_sz = value.nbytes
                else:
                    array_sz = _compute_memory(value)

                dms_sz += array_sz
            return dms_sz

        dms_sz = _compute_memory(self.data)
        self.memory_used = dms_sz
        return dms_sz

    def add_data(self, value, *args):
        """Add value to self.data[arg[0]][...][arg[N-1]]. Requeues self.data[arg[0]] after updating."""
        # def _add_data(keys, val):
        #     val = {keys.pop(): val}
        #     if keys:
        #         val = _add_data(keys, val)
        #     return val

        # # TODO: generalize this to nested dictionaries greater than depth=2...
        # args = list(args)

        # if args[0] in self.data:
        #     # dictionary at depth=1 already defined for key=args[0]
        #     tmp_dict = _add_data(args[1:], value)
        #     self.data[args[0]].update(tmp_dict)
        # else:
        #     # dictionary at depth=1 not defined yet for key=args[0]
        #     try:
        #         tmp_dict = _add_data(args[1:], value)
        #     except IndexError:
        #         # this dictionary only goes to depth=2
        #         tmp_dict = value
        #     finally:
        #         self.data[args[0]] = tmp_dict
        self.data[args[0]] = value

        self.requeue(args[0])
        self.manage_memory()

    def get_data(self, *args):
        """Retrieve NumPy ndarray from self.data according to provided sequence of keys."""
        tmp = self.data

        for key in args:
            if isinstance(tmp, np.ndarray):
                print(
                    ">>> Warning: Already reached end of data tree. Too many arguments provided."
                )
                print("keys provided: {0}".format(args))
                print("current key: {0}".format(key))
                break
            else:
                try:
                    tmp = tmp[key]
                except KeyError:
                    logging.info("DMS: Data not yet in DMS, loading...")
                    raise (KeyError("KeyError when retrieving: {0}".format(key)))

        self.requeue(args[0])
        logging.info("DMS: Data located in DMS, retrieving...")
        return tmp



"""Market data processing tools."""
# from __future__ import absolute_import
# from xlrd.biffh import XLRDError

import pandas as pd
import numpy as np
import calendar
import logging


def read_ercot_da_spp(fname, month, settlement_point):
    """
    Read the day-ahead Ercot historical energy prices from file and return the NumPy ndarray corresponding to the hourly price at settlement_point.

    :param fname: string giving location of DAM SPPs (day-ahead market settlement point prices, a.k.a LMPs) file
    :type fname: str
    :param month: which month to read data for; (int) [1, 12] OR (str) ['1', '12']
    :type month: int or str
    :param settlement_point: string giving settlement point name (hub or load zone)
    :type settlement_point: str
    :return spp_da: NumPy ndarray with SPPs for month and settlement point
    :rtype: NumPy ndarray
    """
    spp_da = np.array([])

    # if month is provided as an int, map it to the correct calendar month
    if isinstance(month, int):
        month_abbr = calendar.month_abbr(month)
    elif isinstance(month, str):
        month_ix = int(month)
        month_abbr = calendar.month_abbr[month_ix]

    # Retrieve the correct worksheet for the month.
    wkbk = pd.ExcelFile(fname)
    wkbk_sheetnames = [name[:3] for name in wkbk.sheet_names]

    try:
        wkst_ix = wkbk_sheetnames.index(month_abbr)
    except ValueError:
        # The worksheet for the requested month does not exist.
        logging.warning(
            """read_ercot_da_spp: Could not load data (the specified month of data could not be found in the given file), returning empty array.
            (got {fname},{month}, {settlement_point})""".format(
                fname=fname, month=month, settlement_point=settlement_point
            )
        )
        return spp_da
    else:
        df = wkbk.parse(wkst_ix)

        # filter DataFrame by settlement_point and extract series
        df0 = df.loc[df["Settlement Point"] == settlement_point]
        df1 = df0["Settlement Point Price"]

        # convert to NumPy array and remove NaN
        spp_da = df1.astype("float").values
        spp_da = spp_da[~np.isnan(spp_da)]

    return spp_da


def read_ercot_da_ccp(fname, month):
    """
    Read the day-ahead Ercot capacity clearing prices from file and returns NumPy ndarrays corresponding to the hourly regdn and regup prices.

    :param fname:
    :param month: which month to read data for; (int) [1, 12] OR (str) ['1', '12']

    :param fname: string giving location of DAM CCPs file
    :type fname: str
    :param month: which month to read data for; (int) [1, 12] OR (str) ['1', '12']
    :type month: int or str
    :return: NumPy ndarrays with regdn and regup CCPs for month
    :rtype: Numpy ndarrays
    """
    regdn = np.array([])
    regup = np.array([])

    # if month is provided as an int, map it to the correct calendar month
    if isinstance(month, int):
        month_ix = month
    elif isinstance(month, str):
        month_ix = int(month)

    # Read .csv and generate a Series for the month from "Delivery Date" column.
    df = pd.read_csv(fname, low_memory=False)
    series_month = pd.to_datetime(df["Delivery Date"]).dt.month

    # Filter DataFrame by month.
    df1 = df.loc[series_month == month_ix]

    if len(df1) > 0:
        regdn = df1["REGDN"].astype("float").values
        regdn = regdn[~np.isnan(regdn)]

        try:
            regup = (
                df1["REGUP "].astype("float").values
            )  # why is there an extra space in the key
        except KeyError:
            regup = df1["REGUP"].astype("float").values
        finally:
            regup = regup[~np.isnan(regup)]
    else:
        logging.warning(
            "read_ercot_da_ccp: No data matching input parameters found, returning empty array. (got {fname}, {month})".format(
                fname=fname, month=month
            )
        )

    return regdn, regup


"""Objective and Contraints Class."""

from __future__ import absolute_import
import pyomo.environ as pom
import math


class ExpressionsBlock:
    """Create blocks for objective and constraint functions and assigns them to the Pyomo model.

    Note-
    1. Time step is 1 hour
    2. The state of charge is actually the state of energy in MWh except for the intial and final state of charge.
    3. The intial and final state of charge are in percent.
    """

    def __init__(self, market_type):
        self._market_type = market_type

    @property
    def market_type(self):
        """Market to create blocks for."""
        return self._market_type

    @market_type.setter
    def market_type(self, value):
        self._market_type = value

    def set_expressions(self, model):
        """Generate the objective and constraint expressions for model."""
        block_obj = pom.Block()
        model.objectives = block_obj

        block_con = pom.Block()
        model.constraints = block_con

        if self.market_type == "arbitrage":
            self._objective_arb(block_obj)
            self._constraints_arb(block_con)
        elif self.market_type == "pfp":
            self._objective_miso_pfp(block_obj)
            self._constraints_miso_pfp(block_con)
        else:
            raise ValueError("Invalid market type specified!")

    def _objective_arb(self, block):
        eq_objective_arb(block)

    def _constraints_arb(self, block):
        eq_stateofcharge_arb(block)
        eq_stateofcharge_initial(block)
        eq_stateofcharge_final(block)
        ineq_stateofcharge_minimum(block)
        ineq_stateofcharge_maximum(block)
        ineq_power_limit(block)

    def _objective_miso_pfp(self, block):
        eq_objective_miso_pfp(block)

    def _constraints_miso_pfp(self, block):
        eq_stateofcharge_miso_pfp(block)
        eq_stateofcharge_initial(block)
        eq_stateofcharge_final(block)
        ineq_stateofcharge_minimum_reserve_oneprod(block)
        ineq_stateofcharge_maximum_reserve_oneprod(block)
        ineq_power_limit_oneprod(block)

#############################
# Arbitrage only ############
#############################


def eq_objective_arb(m):
    """Formulate net revenue over the time horizon for arbitrage only."""
    mp = m.parent_block()

    _expr = sum(
        (mp.price_electricity[t] * mp.q_d[t] - mp.price_electricity[t] * mp.q_r[t])
        * math.e ** (-t * mp.R)
        for t in mp.time
    )

    mp.objective_expr += _expr
    m.objective_rt = pom.Expression(expr=_expr)


def eq_stateofcharge_arb(m):
    """Formulate state of charge for storage device in participating in arbitrage only."""
    mp = m.parent_block()

    def _eq_stateofcharge_arb(_m, t):
        return (
            mp.Self_discharge_efficiency * mp.s[t]
            + mp.Round_trip_efficiency * mp.q_r[t]
            - mp.q_d[t]
            == mp.s[t + 1]
        )

    m.stateofcharge = pom.Constraint(mp.time, rule=_eq_stateofcharge_arb)


def ineq_power_limit(m):
    """Limit the energy charged and discharged at each timestep to the device power rating."""
    mp = m.parent_block()

    def _ineq_power_limit(_m, t):
        return mp.Power_rating >= mp.q_r[t] + mp.q_d[t]

    m.power_limit = pom.Constraint(mp.time, rule=_ineq_power_limit)

#############################
# MISO Pay-for-Performance ##
#############################


def eq_objective_miso_pfp(m):
    """Formualte net revenue over the time horizon for pay-for-performance in MISO regulation market."""
    mp = m.parent_block()
    _expr = sum(
        (
            mp.price_electricity[t] * mp.q_d[t]
            - mp.price_electricity[t] * mp.q_r[t]
            + (1 + mp.Make_whole)
            * mp.q_reg[t]
            * mp.perf_score[t]
            * mp.price_regulation[t]
        )
        * math.e ** (-t * mp.R)
        for t in mp.time
    )

    mp.objective_expr += _expr
    m.objective_rt = pom.Expression(expr=_expr)


def eq_stateofcharge_miso_pfp(m):
    """Formulate state of charge for storage device in MISO market."""
    mp = m.parent_block()

    def _eq_stateofcharge_miso_pfp(_m, t):
        return (
            mp.Self_discharge_efficiency * mp.s[t]
            + mp.Round_trip_efficiency * mp.q_r[t]
            - mp.q_d[t]
            + mp.Round_trip_efficiency * mp.fraction_reg_down[t] * mp.q_reg[t]
            - mp.fraction_reg_up[t] * mp.q_reg[t]
            == mp.s[t + 1]
        )

    m.stateofcharge = pom.Constraint(mp.time, rule=_eq_stateofcharge_miso_pfp)

##################################
#          Generic              ##
#     Equality Constraints      ##
##################################


def eq_stateofcharge_initial(m):
    """Set the initial state of energy equal to the inital state of charge time the energy capacity."""
    mp = m.parent_block()

    def _eq_stateofcharge_initial(_m, t):
        if not t == 0:
            return pom.Constraint.Skip
        else:
            return mp.s[t] == mp.State_of_charge_init * mp.Energy_capacity

    m.stateofcharge_initial = pom.Constraint(
        mp.soc_time, rule=_eq_stateofcharge_initial
    )


def eq_stateofcharge_final(m):
    """Set the final state of energy equal to the inital state of energy."""
    mp = m.parent_block()

    def _eq_stateofcharge_final(_m, t):
        if not t == mp.soc_time[-1]:
            return pom.Constraint.Skip
        else:
            return mp.s[t] == mp.State_of_charge_init * mp.Energy_capacity

    m.stateofcharge_final = pom.Constraint(mp.soc_time, rule=_eq_stateofcharge_final)


##################################
#          Generic              ##
#   Inequality Constraints      ##
##################################


def ineq_stateofcharge_minimum(m):
    """Keep the state of energy of the energy storage device above the minimum at any given time."""
    mp = m.parent_block()

    def _ineq_stateofcharge_minimum(_m, t):
        return mp.s[t] >= mp.State_of_charge_min * mp.Energy_capacity

    m.stateofcharge_minimum = pom.Constraint(
        mp.soc_time, rule=_ineq_stateofcharge_minimum
    )


def ineq_stateofcharge_maximum(m):
    """Keep the state of energy of the energy storage device below the maximum at any given time."""
    mp = m.parent_block()

    def _ineq_stateofcharge_maximum(_m, t):
        return mp.s[t] <= mp.State_of_charge_max * mp.Energy_capacity

    m.stateofcharge_maximum = pom.Constraint(
        mp.soc_time, rule=_ineq_stateofcharge_maximum
    )


def ineq_stateofcharge_minimum_reserve_oneprod(m):
    """Keep the state of energy of the energy storage device above (the minimum + a minimum reserve amount) at any given time.

    Note-
    1. The minimum state of charge is defined by the storage operator following operational/safety/longevity guidelines.
    2. The minimum reserve amount is required by the market so that a resource has enough energy ready to be called upon for discharging.
    """
    mp = m.parent_block()

    def _ineq_stateofcharge_minimum(_m, t):
        return (
            mp.s[t + 1]
            >= mp.Reserve_reg_min  # The number of hrs that storage device can at least discharge at q_reg rate.
            * mp.q_reg[t]
            + mp.State_of_charge_min * mp.Energy_capacity
        )

    m.stateofcharge_minimum = pom.Constraint(mp.time, rule=_ineq_stateofcharge_minimum)


def ineq_stateofcharge_maximum_reserve_oneprod(m):
    """Keep the state of energy of the energy storage device below (the maximum + a headroom) at any given time, accounting for penalty aversion parameters.

    Note-
    1. The maximum state of charge is defined by the storage operator following operational/safety/longevity guidelines.
    2. The headroom is required by the market so that a resource has enough space ready to be called upon for charging.
    """
    mp = m.parent_block()

    def _ineq_stateofcharge_maximum(_m, t):
        return (
            mp.s[t + 1]
            <= mp.State_of_charge_max * mp.Energy_capacity
            - mp.Round_trip_efficiency
            * mp.Reserve_reg_max  # The number of hrs that storage device can at least charge at q_reg rate.
            * mp.q_reg[t]
        )

    m.stateofcharge_maximum = pom.Constraint(mp.time, rule=_ineq_stateofcharge_maximum)

def ineq_power_limit_oneprod(m):
    """Limit the energy charged and discharged at each timestep to the device power rating."""
    mp = m.parent_block()

    def _ineq_power_limit_oneprod(_m, t):
        return mp.Power_rating >= mp.q_r[t] + mp.q_d[t] + mp.q_reg[t]

    m.power_limit = pom.Constraint(mp.time, rule=_ineq_power_limit_oneprod)


from __future__ import absolute_import

import os
import textwrap

import matplotlib as mpl
import matplotlib.pyplot as plt

from kivy.properties import ObjectProperty, BooleanProperty
from kivy.uix.modalview import ModalView
from kivy.core.window import Window

from es_gui.resources.widgets.common import ResultsViewer


class ValuationResultsViewer(ResultsViewer):
    """The screen for displaying plots inside the application or exporting results."""
    current_fig = ObjectProperty()
    current_ax = ObjectProperty()

    def __init__(self, **kwargs):
        super(ValuationResultsViewer, self).__init__(**kwargs)

        self.dfs = {}

        self.time_selector.start_time.bind(on_text_validate=self.draw_figure)
        self.time_selector.end_time.bind(on_text_validate=self.draw_figure)

    def on_pre_enter(self):
        """Updates the navigation bar's title."""
        ab = self.manager.nav_bar
        # ab.build_valuation_results_nav_bar()

        #Window.bind(on_key_down=self._on_keyboard_down)

        self._update_toolbar()

        self.rv.data = self.manager.get_screen('valuation_home').handler.get_solved_ops()
        # self.run_selector.rv.data = self.manager.get_screen('valuation_home').handler.solved_ops

    def _update_toolbar(self, *args):
        """Updates the data viewing toolbar based on selections."""
        super(ValuationResultsViewer, self)._update_toolbar(self)

        vars_list = ['revenue', 'state of charge', 'arbitrage activity', 'regulation capacity offered', 'price of electricity', 'price of electricity (box)']

        self.vars_button.values = vars_list

    def draw_figure(self, *args):
        self._update_selection()

        if not self._validate_inputs():
            return

        start_time, end_time = self.time_selector.get_inputs()
        plot_type = self.vars_button.text

        results = self.dfs

        if plot_type in ['price of electricity (box)']:
            self._reinit_graph(has_legend=False)
        else:
            self._reinit_graph(has_legend=True)

        fig = self.current_fig
        ax = self.current_ax
        ax.clear()

        plt.xticks(rotation=0)
        plt.grid(True)

        if plot_type == 'arbitrage activity':
            for key in results:
                df = results[key]
                ax.plot((df['q_r'] - df['q_d'])[start_time:end_time], drawstyle='steps-post', label=textwrap.fill(key, 50))

            ax.set_ylabel('MWh')
            ax.set_xlabel('ending hour')
            ax.set_title('Energy Charged/Discharged (MWh)')
        elif plot_type == 'regulation capacity offered':
            for key in results:
                df = results[key]

                if (df['q_reg'] != 0).any():
                    ax.plot(df['q_reg'][start_time:end_time], drawstyle='steps-post', label=textwrap.fill(key, 50))
                else:
                    ax.plot((df['q_rd'] - df['q_ru'])[start_time:end_time], drawstyle='steps-post', label=textwrap.fill(key, 50))

            ax.set_ylabel('MWh')
            ax.set_xlabel('ending hour')
            ax.set_title('Regulation Capacity Offered (MWh)')
        elif plot_type == 'revenue':
            for key in results:
                df = results[key]
                ax.plot(df['revenue'][start_time:end_time], drawstyle='steps-post', label=textwrap.fill(key, 50))

                ax.set_ylabel('$')
                ax.set_xlabel('ending hour')
                ax.set_title('Cumulative Revenue Generated ($)')
                ax.get_yaxis().set_major_formatter(
                    mpl.ticker.FuncFormatter(lambda x, p: format(int(x), ',')))
        elif plot_type == 'state of charge':
            for key in results:
                df = results[key]
                ax.plot(df['state of charge'][start_time:end_time], drawstyle='steps-post', label=textwrap.fill(key, 50))

                ax.set_ylabel('MWh')
                ax.set_xlabel('ending hour')
                ax.set_title('State of Charge (MWh)')
        elif plot_type == 'price of electricity':
            for key in results:
                df = results[key]
                ax.plot(df['price of electricity'][start_time:end_time], drawstyle='steps-post', label=textwrap.fill(key, 50))

                ax.set_ylabel('$/MWh')
                ax.set_xlabel('ending hour')
                ax.set_title('Price of Electricity ($/MWh)')
        elif plot_type == 'price of electricity (box)':
                ax.boxplot([results[key]['price of electricity'][start_time:end_time] for key in results], labels=[textwrap.fill(' '.join(key.split()[:3]), 8) for key in results])

                ax.set_ylabel('$/MWh')
                ax.set_title('Price of Electricity ($/MWh)')
        else:
            for key in results:
                df = results[key]
                ax.plot(df[plot_type][start_time:end_time], drawstyle='steps-post', label=textwrap.fill(key, 50))

                ax.set_title(plot_type)

        if plot_type in ['price of electricity (box)']:
            pass
        else:
            ax.legend(bbox_to_anchor=(1.02, 0.5), loc="center left", borderaxespad=0, shadow=False, labelspacing=1.8)

        self.plotbox.children[0].draw()

    def export_png(self):
        """Exports currently displayed figure to .png file in specified location."""
        outdir_root = os.path.join('results', 'valuation', 'plots')

        super(ValuationResultsViewer, self).export_png(outdir_root)

    def export_csv(self):
        """Exports selected DataFrames to .csv files in specified location."""
        outdir_root = os.path.join('results', 'valuation', 'csv')

        super(ValuationResultsViewer, self).export_csv(outdir_root)


class RunSelector(ModalView):
    """
    A ModalView used for selecting which saved runs to view or export results from.
    """
    pass



import logging
import os
from datetime import datetime

import pandas as pd

from kivy.uix.modalview import ModalView
from kivy.uix.filechooser import FileChooserListView
from kivy.uix.screenmanager import Screen, ScreenManager
from kivy.properties import StringProperty, BooleanProperty

from es_gui.resources.widgets.common import RecycleViewRow, WarningPopup


class DataImporterFileChooser(FileChooserListView):
    """FileChooserListView for selecting file to import.
    """
    def __init__(self, *args, **kwargs):
        super(DataImporterFileChooser, self).__init__(**kwargs)

        self.filters = ['*.csv',]
        self.multiselect = False

    def on_submit(self, selection, touch):
        self.host_view.file_selected()


class DataImporterFileChooserScreen(Screen):
    """DataImporter screen for selecting which file to import.
    """
    def __init__(self, *args, **kwargs):
        super(DataImporterFileChooserScreen, self).__init__(**kwargs)

        DataImporterFileChooser.host_view = self

    def file_selected(self):
        self._validate_file_selected()
    
    def _validate_file_selected(self):
        try:
            file_selected = self.filechooser.selection[0]
        except IndexError:
            pass
        else:
            file_selected_ext = file_selected.split('.')[-1]

            if file_selected_ext == 'csv':
                logging.info('DataImporter: {0} is a valid csv file.'.format(file_selected))
                self.manager.file_selected = file_selected

                self.manager.current = self.manager.next()
            else:
                logging.error('DataImporter: {0} is not a valid csv file.'.format(file_selected))


class DataImporterFormatAnalyzerScreen(Screen):
    """DataImporter screen for selecting which column to use and completing the import process.
    """
    datetime_column = StringProperty("")
    data_column = StringProperty("")
    has_selections = BooleanProperty(False)

    def __init__(self, **kwargs):
        super(DataImporterFormatAnalyzerScreen, self).__init__(**kwargs)

        DataColumnRecycleViewRow.data_analyzer_screen = self

    def on_pre_enter(self):
        """Populates the recycle view with column names based on selected file.
        """
        file_selected = self.manager.file_selected
        self.file_selected_df = pd.read_csv(file_selected)

        column_options = [{'name': column} 
        for column in self.file_selected_df.columns
        ]

        self.data_col_rv.data = column_options
        self.data_col_rv.unfiltered_data = column_options
    
    def on_data_column(self, instance, value):
        """Checks if data column has been specified.
        """
        logging.info('DataImporter: Data column changed to {0}.'.format(self.data_column))

        self.has_selections = (self.data_column != "")

    def on_has_selections(self, instance, value):
        """Activates the import button if both columns have been specified.
        """
        if value:
            logging.info("DataImporter: All selections have been made.")
        
        self.import_button.disabled = not value
    
    def finalize_selections(self):
        """Validates the specified data based on the selected columns using a validation function.

        Returns
        -------
        Pandas DataFrame
            Dataframe with the data series
        
        str
            Name of the data column
        """
        try:
            import_df, data_column_name = self._validate_columns_selected()
        except ValueError as e:
            exception_popup = WarningPopup()
            exception_popup.popup_text.text = e.args[0]
            exception_popup.open()
        else:
            logging.info("DataImporter: Data format validation completed without issues.")

            completion_popup = self.manager.completion_popup
            completion_popup.open()

            return import_df, data_column_name
    
    def _validate_columns_selected(self):
        """Validates the data using a validation function.

        Raises
        ------
        ValueError
            If validation fails based on ``self.validation_function``
        """
        self.data_validation_function(self.file_selected_df, self.data_column)

        return self.file_selected_df, self.data_column
    
    def get_selections(self):
        """Returns the data import results.

        Returns
        -------
        Pandas DataFrame
            DataFrame containing a single Series for the data
        
        str
            Name of the data column
        """
        import_df, data_column_name = self._validate_columns_selected()

        return import_df, data_column_name


class DataColumnRecycleViewRow(RecycleViewRow):
    """The representation widget for column names in the data column selector RecycleView."""
    data_analyzer_screen = None

    def on_touch_down(self, touch):
        """Add selection on touch down."""
        if super(DataColumnRecycleViewRow, self).on_touch_down(touch):
            return True
        if self.collide_point(*touch.pos) and self.selectable:
            return self.parent.select_with_touch(self.index, touch)

    def apply_selection(self, rv, index, is_selected):
        """Respond to the selection of items in the view."""
        self.selected = is_selected

        if is_selected:
            self.data_analyzer_screen.data_column = rv.data[self.index]['name']


class DataImporter(ModalView):
    """A ModalView with a series of prompts for importing time series data from a csv file.
    
    Parameters
    ----------
    write_directory : str
        Path of directory where the imported data will be written

    write_function : func
        Function describing how to write the imported data to a persistent file
    
    chooser_description : str
        Description displayed on the DataImporter file chooser screen
    
    format_description : str
        Description displayed on the DataImporter format analyzer screen
    
    data_validation_function : str
        Function used to validate the selected data

    Notes
    -----
    ``write_function`` should handle saving the persistent object (e.g., csv file) to disk.
    ``data_validation_function`` should raise a ValueError to indicate failing validation with a relevant reason why it failed.    
    """
    def __init__(self, write_directory=None, write_function=None, chooser_description=None, format_description=None, data_validation_function=None, **kwargs):
        super(DataImporter, self).__init__(**kwargs)

        if write_directory is None:
            self.write_directory = ""
        else:
            self.write_directory = write_directory
        
        if write_function is None:
            def _write_time_series_csv(fname, dataframe):
                """Writes a generic time series dataframe to a two-column csv. 
                
                The data is inferred to be at an hourly time resolution for one standard year.

                Parameters
                ----------
                fname : str
                    Name of the file to be saved without an extension
                dataframe : Pandas DataFrame
                    DataFrame containing a single Series of the data
                
                Returns
                -------
                str
                    The save destination of the resulting file.
                """
                save_destination = os.path.join(self.write_directory, fname + ".csv")

                data_column_name = dataframe.columns[0]

                datetime_start = datetime(2019, 1, 1, 0)
                hour_range = pd.date_range(start=datetime_start, periods=len(dataframe), freq="H")
                dataframe["DateTime"] = hour_range

                dataframe[["DateTime", data_column_name]].to_csv(save_destination, index=False)
                
                return save_destination
            
            self.write_function = _write_time_series_csv
        else:
            self.write_function = write_function

        if chooser_description is None:
            self.chooser_description = "Select a .csv file to import data from."
        else:
            self.chooser_description = chooser_description
        
        file_chooser_screen = self.screen_manager.get_screen("FileChooser")
        file_chooser_screen.file_chooser_body_text.text = self.chooser_description

        if format_description is None:
            self.format_description = "Specify the data column."
        else:
            self.format_description = format_description

        format_analyzer_screen = self.screen_manager.get_screen("FormatAnalyzer")
        format_analyzer_screen.format_analyzer_body_text.text = self.format_description
    
        if data_validation_function is None:
            def _default_data_validation_function(dataframe, data_column_name):
                if len(dataframe) != 8760:
                    raise ValueError("The length of the time series must be 8760 (got {0}).".format(len(dataframe)))
                
                data_column = dataframe[data_column_name]

                try:
                    data_column.astype("float")
                except ValueError:
                    raise ValueError("The selected data column could not be interpeted as numeric float values.")


            self.data_validation_function = _default_data_validation_function
        else:
            self.data_validation_function = data_validation_function

        # Bind DataImporter dismissal to successful data import.
        completion_popup = WarningPopup()
        completion_popup.title = "Success!"
        completion_popup.popup_text.text = "Data successfully imported."
        completion_popup.bind(on_dismiss=self.dismiss)

        self.screen_manager.completion_popup = completion_popup

    @property
    def write_directory(self):
        """The directory where the imported time series data will be saved."""
        return self._write_directory
    
    @write_directory.setter
    def write_directory(self, value):
        self._write_directory = value
    
    @property
    def write_function(self):
        """The function used to write the imported data to disk."""
        return self._write_function
    
    @write_function.setter
    def write_function(self, value):
        self._write_function = value
    
    @property
    def chooser_description(self):
        """Description displayed on the file chooser screen."""
        return self._chooser_description
    
    @chooser_description.setter
    def chooser_description(self, value):
        self._chooser_description = value
        file_chooser_screen = self.screen_manager.get_screen("FileChooser")
        file_chooser_screen.file_chooser_body_text.text = self.chooser_description
    
    @property
    def format_description(self):
        """Description displayed on the format analyzer screen."""
        return self._format_description
    
    @format_description.setter
    def format_description(self, value):
        self._format_description = value
        format_analyzer_screen = self.screen_manager.get_screen("FormatAnalyzer")
        format_analyzer_screen.format_analyzer_body_text.text = self.format_description
    
    @property
    def data_validation_function(self):
        """"The function used to validate the format of the selected data."""
        return self._data_validation_function
    
    @data_validation_function.setter
    def data_validation_function(self, value):
        self._data_validation_function = value
        format_analyzer_screen = self.screen_manager.get_screen("FormatAnalyzer")
        format_analyzer_screen.data_validation_function = self.data_validation_function
    
    def get_import_selections(self):
        """Returns the destination of the processed imported data.

        This method pulls the selections from the DataImporter prompts. 
        Using the selected data, it writes a formatted version of the data to disk according to specification.
        
        Returns
        -------
        str
            The save destination of the file with the imported data.
        """
        imported_filename = self.screen_manager.file_selected
        import_df, data_column_name = self.screen_manager.get_screen("FormatAnalyzer").get_selections()

        # Write imported time series data to write_directory.
        os.makedirs(self.write_directory, exist_ok=True)

        # Strip non-alphanumeric chars from data column name.
        delchars = ''.join(c for c in map(chr, range(256)) if not c.isalnum())

        generated_save_name = "_".join([os.path.split(imported_filename)[-1][:-4], data_column_name.translate({ord(i): None for i in delchars})])
        dataframe = import_df[[data_column_name]]

        save_destination = self.write_function(generated_save_name, dataframe)
        logging.info('DataImporter: Selected time series saved to {0}'.format(save_destination))

        return save_destination
        


"""Market data processing tools."""
# from __future__ import absolute_import
# from xlrd.biffh import XLRDError

import pandas as pd
import numpy as np
import os
import logging


def read_caiso_data(fpath, year, month, nodeid):
    """Read the historical LMP, regulation up/down and regulation mileage up/down and return NumPy ndarrays for those three prices.

    :param fpath: The path to the root of the PJM data directory
    :type fpath: str
    :param year: Year of data to read
    :type year: int or str
    :param month: Month of data to read
    :type month: int or str
    :param nodeid: ID of the node to read
    :type nodeid: str
    :return: daLMP, RegCCP, RegPCP: Hourly LMP and regulation capacity/performance clearing price values.
    :rtype: NumPy ndarrays

    Note-
    1. For CAISO certain prices have:
    _CAISO
    _CAISO_EXP
    _NP26
    _NP26_EXP
    _SP26
    _SP26_EXP

    2. we only use the _CAISO_EXP ones for mileage prices as they are the only ones

    AS_CAISO_EXP_RD_CLR_PRC
    AS_CAISO_EXP_RU_CLR_PRC

    AS_CAISO_EXP_RMD_CLR_PRC
    AS_CAISO_EXP_RMU_CLR_PRC
    """
    daLMP = np.empty([0])
    daREGU = np.empty([0])
    daREGD = np.empty([0])
    daRMU = np.empty([0])
    daRMD = np.empty([0])
    RMU_MM = np.empty([0])
    RMD_MM = np.empty([0])
    RMU_PACC = np.empty([0])
    RMD_PACC = np.empty([0])

    if isinstance(month, str):
        month = int(month)

    if isinstance(year, str):
        year = int(year)

    # Names examples:
    # 201601_dalmp_LAKESID2_7_UNITS-APND.csv
    # 201601_asp.csv
    # 201601_regm.csv
    fnameLMP = "{0:d}{1:02d}_dalmp_{2:s}.csv".format(year, month, nodeid)
    fnameASP = "{0:d}{1:02d}_asp.csv".format(year, month)
    fnameMILEAGE = "{0:d}{1:02d}_regm.csv".format(year, month)

    fname_path_LMP = os.path.join(fpath, "LMP", str(nodeid), str(year), fnameLMP)
    fname_path_ASP = os.path.join(fpath, "ASP", str(year), fnameASP)
    fname_path_MILEAGE = os.path.join(fpath, "MILEAGE", str(year), fnameMILEAGE)

    try:
        dfLMP = pd.read_csv(fname_path_LMP, index_col=False)
        daLMP = dfLMP["LMP"].values
    except FileNotFoundError:
        logging.warning(
            "read_caiso_data: No LMP data matching input parameters found, returning empty array. (got {fname}, {year}, {month}, {nodeid})".format(
                fname=fnameLMP, year=year, month=month, nodeid=nodeid
            )
        )

    try:
        dfASP = pd.read_csv(fname_path_ASP, index_col=False)
        daREGU = dfASP["AS_CAISO_EXP_RU_CLR_PRC"].values
        daREGD = dfASP["AS_CAISO_EXP_RD_CLR_PRC"].values
        daRMU = dfASP["AS_CAISO_EXP_RMU_CLR_PRC"].values
        daRMD = dfASP["AS_CAISO_EXP_RMD_CLR_PRC"].values

    except FileNotFoundError:
        logging.warning(
            "read_caiso_data: No ASP data matching input parameters found, returning empty array. (got {fname}, {year}, {month})".format(
                fname=fnameASP, year=year, month=month
            )
        )

    try:
        dfMIL_ACC = pd.read_csv(fname_path_MILEAGE, index_col=False)
        RMU_MM = dfMIL_ACC["RMU_SYS_MIL_MUL"].values
        RMD_MM = dfMIL_ACC["RMD_SYS_MIL_MUL"].values
        RMU_PACC = dfMIL_ACC["RMU_SYS_PERF_ACC"].values
        RMD_PACC = dfMIL_ACC["RMD_SYS_PERF_ACC"].values

    except FileNotFoundError:
        logging.warning(
            """read_caiso_data: No MILEAGE and PERFORMANCE ACCURACY data matching input parameters found, returning empty array.
            (got {fname}, {year}, {month})""".format(
                fname=fnameMILEAGE, year=year, month=month
            )
        )

    return daLMP, daREGU, daREGD, daRMU, daRMD, RMU_MM, RMD_MM, RMU_PACC, RMD_PACC